# 2023-03-19 Cornelius Kölbel, <cornelius.koelbel@netknights.it>
#            Add new API for service IDs
#
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__doc__ = """The serviceid endpoint allows administrators to manage service ID definitions.
"""
from flask import (Blueprint, request)
from .lib.utils import (getParam, send_result)
from ..lib.log import log_with
from privacyidea.lib.serviceid import get_serviceids, delete_serviceid, set_serviceid
from privacyidea.lib.event import event
from privacyidea.lib.policy import ACTION
from privacyidea.api.lib.prepolicy import prepolicy, check_base_action

from flask import g
import logging


log = logging.getLogger(__name__)


serviceid_blueprint = Blueprint('serviceid_blueprint', __name__)


@serviceid_blueprint.route('/<name>', methods=['POST'])
@prepolicy(check_base_action, request, ACTION.SERVICEID_ADD)
@event("serviceid_add", request, g)
@log_with(log)
def set_serviceid_api(name):
    """
    This call creates a new service ID definition or updates the description
    of an existing service ID.

    Note, that the identifier (name) of the service ID needs to be unique.

    :param name: The unique name of the service ID
    :param description: The description of the service ID definition
    :return:

    **Example request**:

    To create a new serviceid "serviceA" with a description call:

    .. sourcecode:: http

       POST /serviceid/serviceA HTTP/1.1
       Host: example.com
       Accept: application/json
       Content-Length: 26
       Content-Type: application/x-www-form-urlencoded

       description=My cool first service

    **Example response**:

    .. sourcecode:: http

       HTTP/1.1 200 OK
       Content-Type: application/json

       {
          "id": 1,
          "jsonrpc": "2.0",
          "result": {
                    "status": true,
                    "value": 1
          }
          "version": "privacyIDEA unknown"
       }

    """
    param = request.all_data
    description = getParam(param, "description", optional=True)

    r = set_serviceid(name, description)

    g.audit_object.log({'success': r > 0, 'info':  name})
    return send_result(r)


@serviceid_blueprint.route('/<name>', methods=['GET'])
@serviceid_blueprint.route('/', methods=['GET'])
@prepolicy(check_base_action, request, ACTION.SERVICEID_LIST)
@event("serviceid_list", request, g)
@log_with(log)
def get_serviceid_api(name=None):
    """
    This call returns the information for the given service ID.
    If no name is specified, it returns a list of all defined services.

    :return: a json result with a list of services


    **Example request**:

    .. sourcecode:: http

       GET /serviceid/ HTTP/1.1
       Host: example.com
       Accept: application/json

    **Example response**:

    .. sourcecode:: http

        HTTP/1.1 200 OK
        Content-Type: application/json

        {
          "id": 1,
          "jsonrpc": "2.0",
          "result": {
            "status": true,
            "value": {
              "service1": {"description": "1st service"},
              "service2": {"description": "2nd service"}
              }
            }
          },
          "version": "privacyIDEA unknown"
        }
    """
    sids = get_serviceids(name=name)
    g.audit_object.log({"success": True})
    r_serviceids = {}
    for si in sids:
        r_serviceids[si.name] = {"description": si.Description,
                                 "id": si.id}

    return send_result(r_serviceids)


@serviceid_blueprint.route('/<name>', methods=['DELETE'])
@prepolicy(check_base_action, request, ACTION.SERVICEID_DELETE)
@event("serviceid_delete", request, g)
@log_with(log)
def delete_serviceid_api(name=None):
    """
    This call deletes the given service ID definition.

    :param name: The name of the service.

    :return: a json result with value=1 if deleting the service ID was successful

    **Example request**:

    .. sourcecode:: http

       DELETE /serviceid/service1 HTTP/1.1
       Host: example.com
       Accept: application/json

    **Example response**:

    .. sourcecode:: http

       HTTP/1.1 200 OK
       Content-Type: application/json

        {
          "id": 1,
          "jsonrpc": "2.0",
          "result": {
            "status": true,
            "value": 1
          },
          "version": "privacyIDEA unknown"
        }

    """
    ret = delete_serviceid(name)
    g.audit_object.log({"success": True,
                        "info": name})

    return send_result(1)



from alembic import context
from sqlalchemy import engine_from_config, pool
from sqlalchemy.engine.url import make_url
from logging.config import fileConfig
from urllib.parse import quote

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.

config = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
fileConfig(config.config_file_name)

# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
# target_metadata = mymodel.Base.metadata
from flask import current_app


def set_database_url(config):
    url = current_app.config.get('SQLALCHEMY_DATABASE_URI')
    try:
        # In case of MySQL, add ``charset=utf8`` to the parameters (if no charset is set),
        # because this is what Flask-SQLAlchemy does
        if url.startswith("mysql"):
            parsed_url = make_url(url)
            parsed_url = parsed_url.update_query_dict({"charset": "utf8"})
            # We need to quote the password in case it contains special chars
            parsed_url = parsed_url.set(password=quote(parsed_url.password))
            url = str(parsed_url)
    except Exception as exx:
        print(u"Attempted to set charset=utf8 on connection, but failed: {}".format(exx))
    # set_main_option() requires escaped "%" signs in the string
    config.set_main_option('sqlalchemy.url', url.replace('%', '%%'))


set_database_url(config)
target_metadata = current_app.extensions['migrate'].db.metadata

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url = config.get_main_option("sqlalchemy.url")
    context.configure(url=url)

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    # FIX for Postgres updates
    url = config.get_section(config.config_ini_section).get("sqlalchemy.url")
    driver = url.split(":")[0]

    if driver == "postgresql+psycopg2":
        engine = engine_from_config(
                    config.get_section(config.config_ini_section),
                    prefix='sqlalchemy.',
                    isolation_level="AUTOCOMMIT",
                    poolclass=pool.NullPool)
    else:
        engine = engine_from_config(
                    config.get_section(config.config_ini_section),
                    prefix='sqlalchemy.',
                    poolclass=pool.NullPool)

    connection = engine.connect()
    context.configure(
                connection=connection,
                target_metadata=target_metadata,
                compare_type=True,
                render_as_batch=True
                )

    try:
        with context.begin_transaction():
            context.run_migrations()
    except Exception as e:
        print(f"Exception occurred during database upgrade: {e}")
    finally:
        connection.close()


if context.is_offline_mode():
    print("Running offline")
    run_migrations_offline()
else:
    print("Running online")
    run_migrations_online()



# 2020-01-14 Jean-Pierre Höhmann <jean-pierre.hoehmann@netknights.it>
#
# License:  AGPLv3
# Contact:  https://www.privacyidea.org
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This code was originally adapted from the Duo Security implementation of
# WebAuthn for Python.
#
# Copyright (c) 2017 Duo Security, Inc. All rights reserved.
# License:  BSD-3-Clause
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import binascii
import codecs
import hashlib
import json
import logging
import os
import struct

import cbor2
import cryptography.x509
from OpenSSL import crypto
from cryptography import x509
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import constant_time
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicNumbers, SECP256R1, ECDSA
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15, PSS, MGF1
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicNumbers
from cryptography.hazmat.primitives.hashes import SHA256, SHA1
from cryptography.x509 import load_der_x509_certificate
from webauthn import base64url_to_bytes

from privacyidea.lib.tokens.u2f import url_encode, url_decode
from privacyidea.lib.utils import to_bytes, to_unicode

__doc__ = """
Business logic for WebAuthn protocol.

This file implements the server part of the WebAuthn protocol.

This file is tested in tests/test_lib_tokens_webauthn.py
"""

# Default client extensions
#
DEFAULT_CLIENT_EXTENSIONS = {'appid': None}

# Default authenticator extensions
#
DEFAULT_AUTHENTICATOR_EXTENSIONS = {}

log = logging.getLogger(__name__)


class CosePublicKey:
    """
    The indices of the various parameters in a COSE-formatted public key.
    """

    ALG = 3
    X = -2
    Y = -3
    E = -2
    N = -1


class AttestationType:
    """
    Attestation types known to this implementation.
    """

    BASIC = 'Basic'
    ECDAA = 'ECDAA'
    NONE = 'None'
    ATTESTATION_CA = 'AttCA'
    SELF_ATTESTATION = 'Self'


# Only supporting 'None', 'Basic', and 'Self Attestation' attestation types for now.
SUPPORTED_ATTESTATION_TYPES = (
    AttestationType.BASIC,
    AttestationType.NONE,
    AttestationType.SELF_ATTESTATION
)


class AttestationFormat:
    """
    Attestation format identifiers as registered in the IANA WebAuthn attestation statement format identifiers registry.
    """

    PACKED = 'packed'
    TPM = 'tpm'
    ANDROID_KEY = 'android-key'
    ANDROID_SAFETYNET = 'android-safetynet'
    APPLE = 'apple'
    FIDO_U2F = 'fido-u2f'
    NONE = 'none'


# Only supporting 'fido-u2f', 'packed', and 'none' attestation formats for now.
# TODO: implement android, apple and TPM authentication formats
SUPPORTED_ATTESTATION_FORMATS = (
    AttestationFormat.FIDO_U2F,
    AttestationFormat.PACKED,
    AttestationFormat.NONE
)

REGISTERED_ATTESTATION_FORMATS = (
    AttestationFormat.PACKED,
    AttestationFormat.TPM,
    AttestationFormat.ANDROID_KEY,
    AttestationFormat.ANDROID_SAFETYNET,
    AttestationFormat.APPLE,
    AttestationFormat.FIDO_U2F,
    AttestationFormat.NONE
)


class ClientDataType:
    """
    Client data types used by this implementation.
    """

    CREATE = 'webauthn.create'
    GET = 'webauthn.get'


SUPPORTED_CLIENT_DATA_TYPES = (
    ClientDataType.CREATE,
    ClientDataType.GET
)


class CoseAlgorithm:
    """
    IANA-assigned identifiers of supported COSE algorithms.
    """

    ES256 = -7
    PS256 = -37
    RS256 = -257
    RS1 = -65535  # for tests, otherwise unsupported


SUPPORTED_COSE_ALGORITHMS = (
    CoseAlgorithm.ES256,
    CoseAlgorithm.PS256,
    CoseAlgorithm.RS256,
)


class AttestationForm:
    """
    The different forms of attestation.
    """

    NONE = 'none'
    INDIRECT = 'indirect'
    DIRECT = 'direct'


ATTESTATION_FORMS = (
    AttestationForm.NONE,
    AttestationForm.INDIRECT,
    AttestationForm.DIRECT
)


class UserVerificationLevel:
    """
    The different levels of user verification.
    """

    REQUIRED = 'required'
    PREFERRED = 'preferred'
    DISCOURAGED = 'discouraged'


USER_VERIFICATION_LEVELS = (
    UserVerificationLevel.REQUIRED,
    UserVerificationLevel.PREFERRED,
    UserVerificationLevel.DISCOURAGED
)


class AttestationLevel:
    """
    The different levels of attestation requirement.
    """

    TRUSTED = 'trusted'
    UNTRUSTED = 'untrusted'
    NONE = 'none'


ATTESTATION_LEVELS = (
    AttestationLevel.TRUSTED,
    AttestationLevel.UNTRUSTED,
    AttestationLevel.NONE
)

ATTESTATION_REQUIREMENT_LEVEL = {
    AttestationLevel.TRUSTED: {
        'self_attestation_permitted': False,
        'none_attestation_permitted': False
    },
    AttestationLevel.UNTRUSTED: {
        'self_attestation_permitted': True,
        'none_attestation_permitted': False
    },
    AttestationLevel.NONE: {
        'self_attestation_permitted': True,
        'none_attestation_permitted': True
    }
}

ATTESTATION_REQUIREMENT_LEVELS = (
    ATTESTATION_REQUIREMENT_LEVEL[AttestationLevel.TRUSTED],
    ATTESTATION_REQUIREMENT_LEVEL[AttestationLevel.UNTRUSTED],
    ATTESTATION_REQUIREMENT_LEVEL[AttestationLevel.NONE]
)


class AuthenticatorAttachmentType:
    """
    The different types of authenticator attachment.
    """

    PLATFORM = 'platform'
    CROSS_PLATFORM = 'cross-platform'


AUTHENTICATOR_ATTACHMENT_TYPES = (
    AuthenticatorAttachmentType.PLATFORM,
    AuthenticatorAttachmentType.CROSS_PLATFORM
)


class TRANSPORT:
    """
    The standard transports.
    """

    USB = 'usb'
    BLE = 'ble'
    NFC = 'nfc'
    INTERNAL = 'internal'


TRANSPORTS = (
    TRANSPORT.USB,
    TRANSPORT.BLE,
    TRANSPORT.NFC,
    TRANSPORT.INTERNAL,
)


class COSEKeyException(Exception):
    """
    COSE public key invalid or unsupported.
    """

    pass


class AuthenticationRejectedException(Exception):
    """
    The authentication attempt was rejected.
    """

    pass


class RegistrationRejectedException(Exception):
    """
    The registration attempt was rejected.
    """

    pass


class WebAuthnUserDataMissing(Exception):
    """
    The user data is missing.
    """

    pass


class AuthenticatorDataFlags:
    """
    Authenticator data flags:

    https://www.w3.org/TR/webauthn/#authenticator-data
    """

    USER_PRESENT = 1 << 0
    USER_VERIFIED = 1 << 2
    ATTESTATION_DATA_INCLUDED = 1 << 6
    EXTENSION_DATA_INCLUDED = 1 << 7

    def __init__(self, auth_data):
        """
        Create a new AuthenticatorDataFlags object.

        :param auth_data: The authenticator data.
        :type auth_data: basestring
        :return: An AuthenticatorDataFlags object.
        :rtype: AuthenticatorDataFlags
        """

        self.flags = struct.unpack('!B', auth_data[32:33])[0]

    @property
    def user_present(self):
        """
        :return: Whether the user was present.
        :rtype: bool
        """

        return (self.flags & self.USER_PRESENT) == self.USER_PRESENT

    @property
    def user_verified(self):
        """
        :return: Whether the user's identity was verified.
        :rtype: bool
        """

        return (self.flags & self.USER_VERIFIED) == self.USER_VERIFIED

    @property
    def attestation_data_included(self):
        """
        :return: Whether the authenticator response includes attestation information.
        :rtype: bool
        """

        return (self.flags & self.ATTESTATION_DATA_INCLUDED) == self.ATTESTATION_DATA_INCLUDED

    @property
    def extension_data_included(self):
        """
        :return: Whether the authenticator respone included extension data.
        :rtype: bool
        """

        return (self.flags & self.EXTENSION_DATA_INCLUDED) == self.EXTENSION_DATA_INCLUDED


class WebAuthnMakeCredentialOptions:
    """
    Generate the options passed to navigator.credentials.create()
    """

    def __init__(self,
                 challenge,
                 rp_name,
                 rp_id,
                 user_id,
                 user_name,
                 user_display_name,
                 timeout,
                 attestation,
                 user_verification,
                 public_key_credential_algorithms,
                 icon_url=None,
                 authenticator_attachment=None,
                 authenticator_selection_list=None,
                 location=None,
                 credential_ids=None):
        """
        Create a new WebAuthnMakeCredentialOptions object.

        :param challenge: The challenge is a buffer of cryptographically random bytes needed to prevent replay attacks.
        :type challenge: basestring
        :param rp_name: The name of the relying party.
        :type rp_name: basestring
        :param rp_id: The ID of the relying party.
        :type rp_id: basestring
        :param user_id: The ID for the user credential being generated. This is the privacyIDEA token serial.
        :type user_id: basestring
        :param user_name: The username the user logs in with.
        :type user_name: basestring
        :param user_display_name: The human-readable name of the user.
        :type user_display_name: basestring
        :param icon_url: An optional icon url.
        :type icon_url: basestring
        :param timeout: The time (in milliseconds) that the user has to respond to the prompt for registration.
        :type timeout: int
        :param attestation: This option allows to indicate how important the attestation data is to this registration.
        :type attestation: basestring
        :param user_verification: The importance level of verifying the user.
        :type user_verification: basestring
        :param authenticator_attachment: What type of authenticator to register ("platform", or "cross-platform").
        :type authenticator_attachment: basestring
        :param public_key_credential_algorithms: Which algorithms to allow in order of preference.
        :type public_key_credential_algorithms: list of int
        :param authenticator_selection_list: A whitelist of allowed authenticator AAGUIDs
        :type authenticator_selection_list: list of basestring
        :param location: Whether to ask for the inclusion of location information in the attestation.
        :type location: bool
        :param credential_ids: A list of ids that are already enrolled to the user.
        :type credential_ids: list
        :return: A WebAuthnMakeCredentialOptions object.
        :rtype: WebAuthnMakeCredentialOptions
        """

        self.challenge = challenge
        self.rp_name = rp_name
        self.rp_id = rp_id
        self.user_id = user_id
        self.user_name = user_name
        self.user_display_name = user_display_name
        self.icon_url = icon_url
        self.authenticator_selection_list = authenticator_selection_list
        self.location = bool(location)
        self.exclude_credentials = []
        if credential_ids:
            for cred_id in credential_ids:
                self.exclude_credentials.append({
                    "id": cred_id,
                    "type": "public-key",
                    "transports": list(TRANSPORTS)
                })

        attestation = str(attestation).lower()
        if attestation not in ATTESTATION_FORMS:
            raise ValueError('Attestation string must be one of {0!s}'.format(', '.join(ATTESTATION_FORMS)))
        self.attestation = attestation

        if user_verification is not None:
            user_verification = str(user_verification).lower()
            if user_verification not in USER_VERIFICATION_LEVELS:
                raise ValueError('user_verification must be one of {0!s}'.format(', '.join(USER_VERIFICATION_LEVELS)))
        self.user_verification = user_verification

        if authenticator_attachment is not None:
            authenticator_attachment = str(authenticator_attachment).lower()
            if authenticator_attachment not in AUTHENTICATOR_ATTACHMENT_TYPES:
                raise ValueError(
                    'authenticator_attachment must be one of {0!s}'.format(', '.join(AUTHENTICATOR_ATTACHMENT_TYPES)))
        self.authenticator_attachment = authenticator_attachment

        if int(timeout) < 1:
            raise ValueError('timeout must be a positive integer.')
        self.timeout = timeout

        self.public_key_credential_parameters = [
            {
                'alg': i,
                'type': 'public-key'
            }
            for i in public_key_credential_algorithms
        ]

    @property
    def registration_dict(self):
        """
        :return: The publicKeyCredentialCreationOptions dictionary.
        :rtype: dict
        """
        registration_dict = {
            'challenge': self.challenge,
            'rp': {
                'name': self.rp_name,
                'id': self.rp_id
            },
            'user': {
                'id': self.user_id,
                'name': self.user_name,
                'displayName': self.user_display_name
            },
            'pubKeyCredParams': self.public_key_credential_parameters,
            'authenticatorSelection': {},
            'timeout': self.timeout,
            'excludeCredentials': self.exclude_credentials,
            # Relying parties may use AttestationConveyancePreference to specify their
            # preference regarding attestation conveyance during credential generation.
            'attestation': self.attestation,
            'extensions': {}
        }

        if self.user_verification is not None:
            registration_dict['authenticatorSelection']['userVerification'] = self.user_verification

        if self.authenticator_attachment is not None:
            registration_dict['authenticatorSelection']['authenticatorAttachment'] = self.authenticator_attachment

        if self.icon_url is not None:
            registration_dict['user']['icon'] = self.icon_url

        if self.location:
            registration_dict['extensions']['loc'] = True

        if self.authenticator_selection_list is not None:
            registration_dict['extensions']['authnSel'] = self.authenticator_selection_list

        return registration_dict

    @property
    def json(self):
        """
        :return: The publicKeyCredentialCreationOptions dictionary encoded as JSON.
        :rtype: basestring
        """

        return json.dumps(self.registration_dict)


class WebAuthnAssertionOptions:
    """
    Generate the options passed to navigator.credentials.get()
    """

    def __init__(self,
                 challenge,
                 webauthn_user,
                 transports,
                 user_verification_requirement,
                 timeout):
        """
        Create a new WebAuthnAssertionOptions object.

        :param challenge: The challenge is a buffer of cryptographically random bytes needed to prevent replay attacks.
        :type challenge: basestring
        :param webauthn_user: A user cred or a list of user creds to allow authentication with.
        :type webauthn_user: WebAuthnUser or list of WebAuthnUser
        :param transports: Which transports to ask for.
        :type transports: list of basestring
        :param user_verification_requirement: The level of user verification this authentication requires.
        :type user_verification_requirement: basestring
        :param timeout: The time (in milliseconds) that the user has to respond to the prompt for authentication.
        :type timeout: int
        :return: A WebAuthnAssertionOptions object.
        :rtype: WebAuthnAssertionOptions
        """

        self.challenge = challenge
        if not self.challenge:
            raise ValueError('The challenge may not be empty.')

        self.webauthn_users = webauthn_user if isinstance(webauthn_user, list) else [webauthn_user]
        if not self.webauthn_users:
            raise ValueError('webauthn_user may not be empty.')
        for user in self.webauthn_users:
            if not isinstance(user, WebAuthnUser):
                raise ValueError('webauthn_user must be of type WebAuthnUser.')
            if not user.credential_id:
                raise ValueError('user must have a credential_id.')
            if not user.rp_id:
                raise ValueError('user must have a rp_id.')

        if len(set([u.rp_id for u in self.webauthn_users])) != 1:
            raise ValueError('all users must have the same rp_id.')
        self.rp_id = self.webauthn_users[0].rp_id

        self.timeout = timeout
        if int(self.timeout) < 1:
            raise ValueError('timeout must be a positive integer.')

        self.transports = transports
        if not self.transports:
            raise ValueError('transports may not be empty.')

        self.user_verification_requirement = str(user_verification_requirement).lower()
        if self.user_verification_requirement not in USER_VERIFICATION_LEVELS:
            raise ValueError(
                'user_verification_requirement must be one of {0!s}'.format(', '.join(USER_VERIFICATION_LEVELS)))

    @property
    def assertion_dict(self):
        """
        :return: The publicKeyCredentialRequestOptions dictionary.
        :rtype: dict
        """

        return {
            'challenge': self.challenge,
            'allowCredentials': [
                {
                    'type': 'public-key',
                    'id': user.credential_id,
                    'transports': self.transports
                }
                for user in self.webauthn_users
            ],
            'rpId': self.rp_id,
            'userVerification': self.user_verification_requirement,
            'timeout': self.timeout
        }

    @property
    def json(self):
        """
        :return: The publicKeyCredentialRequestOptions dictionary encoded as JSON.
        :rtype: basestring
        """

        return json.dumps(self.assertion_dict)


class WebAuthnUser:
    """
    A single WebAuthn user credential.
    """

    def __init__(self,
                 user_id,
                 user_name,
                 user_display_name,
                 icon_url,
                 credential_id,
                 public_key,
                 sign_count,
                 rp_id):
        """
        Create a new WebAuthnUser object.

        :param user_id: The ID for the user credential being stored. This is the privacyIDEA token serial.
        :type user_id: basestring
        :param user_name: The username the user logs in with.
        :type user_name: basestring
        :param user_display_name: The human-readable name of the user.
        :type user_display_name: basestring
        :param icon_url: An optional icon url.
        :type icon_url: basestring
        :param credential_id: The ID of the credential.
        :type credential_id: basestring
        :param public_key: The credential public key.
        :type public_key: basestring
        :param sign_count: The signature counter value.
        :type sign_count: int
        :param rp_id: The ID of the relying party associated with this credential.
        :type rp_id: basestring
        :return: A WebAuthnUser object.
        :rtype: WebAuthnUser
        """

        if not credential_id:
            raise WebAuthnUserDataMissing("credential_id missing")
        if not public_key:
            raise WebAuthnUserDataMissing("public_key missing")
        if not rp_id:
            raise WebAuthnUserDataMissing("rp_id missing")

        self.user_id = user_id
        self.user_name = user_name
        self.user_display_name = user_display_name
        self.icon_url = icon_url
        self.credential_id = credential_id
        self.public_key = public_key
        self.sign_count = sign_count
        self.rp_id = rp_id

    def __str__(self):
        return '{!r} ({}, {}, {})'.format(self.user_id, self.user_name,
                                          self.user_display_name, self.sign_count)


class WebAuthnCredential:
    """
    A single WebAuthn credential.
    """

    def __init__(self,
                 rp_id,
                 origin,
                 aaguid,
                 credential_id,
                 public_key,
                 sign_count,
                 attestation_level,
                 attestation_cert=None):
        """
        Create a new WebAuthnCredential object.

        :param rp_id: The relying party ID.
        :type rp_id: basestring
        :param origin: The origin of the user the credential is for.
        :type origin: basestring
        :param aaguid: The aaguid of the token used to create the credential
        :type aaguid: bytes
        :param credential_id: The ID of the credential.
        :type credential_id: basestring or bytes
        :param public_key: The public key of the credential.
        :type public_key: basestring or bytes
        :param sign_count: The signature count.
        :type sign_count: int
        :param attestation_level: The level of attestation that was provided for this credential.
        :type attestation_level: basestring
        :param attestation_cert: The attestation certificate, if any.
        :type attestation_cert: Certificate
        :return: A WebAuthnCredential
        :rtype: WebAuthnCredential
        """

        self.rp_id = rp_id
        self.origin = origin
        self.aaguid = aaguid
        self.credential_id = to_bytes(credential_id)
        self.public_key = to_bytes(public_key)
        self.sign_count = sign_count
        self.attestation_cert = attestation_cert

        attestation_level = str(attestation_level).lower()
        if attestation_level not in ATTESTATION_LEVELS:
            raise ValueError('Attestation level must be one of {0!s}'.format(', '.join(ATTESTATION_LEVELS)))
        self.attestation_level = attestation_level

    @property
    def has_signed_attestation(self):
        """
        :return: Whether this credential was created with a signed attestation.
        :rtype: bool
        """

        return not ATTESTATION_REQUIREMENT_LEVEL[self.attestation_level]['none_attestation_permitted']

    @property
    def has_trusted_attestation(self):
        """
        :return: Whether this credential was created with an attestation signed by a trusted root.
        :rtype: bool
        """

        return not ATTESTATION_REQUIREMENT_LEVEL[self.attestation_level]['self_attestation_permitted']

    def __str__(self):
        return '{!r} ({}, {}, {})'.format(self.credential_id, self.rp_id,
                                          self.origin, self.sign_count)


class WebAuthnRegistrationResponse:
    """
    The WebAuthn registration response containing all information needed to verify the registration ceremony.
    """

    def __init__(self,
                 rp_id,
                 origin,
                 registration_response,
                 challenge,
                 attestation_requirement_level,
                 trust_anchor_dir=None,
                 uv_required=False,
                 expected_registration_client_extensions=None,
                 expected_registration_authenticator_extensions=None):
        """
        Create a new WebAuthnRegistrationResponse object.

        :param rp_id: The Relying party id.
        :type rp_id: basestring
        :param origin: The origin of the user.
        :type origin: basestring
        :param registration_response: The registration response containing the client data and attestation.
        :type registration_response: dict
        :param challenge: The challenge that was sent to the client.
        :type challenge: basestring
        :param attestation_requirement_level: Which level of attestation to allow without failing the ceremony.
        :type attestation_requirement_level: dict
        :param trust_anchor_dir: The path to the directory containing the trust anchors
        :type trust_anchor_dir: basestring
        :param uv_required: Whether user verification is required.
        :type uv_required: bool
        :param expected_registration_client_extensions: A dict whose keys indicate which client extensions are expected.
        :type expected_registration_client_extensions: dict
        :param expected_registration_authenticator_extensions: A dict whose keys indicate which auth exts to expect.
        :return: A WebAuthnRegistrationResponse object.
        :rtype: WebAuthnRegistrationResponse
        """

        self.rp_id = rp_id
        self.origin = origin
        self.registration_response = registration_response
        self.challenge = challenge
        self.trust_anchor_dir = trust_anchor_dir
        self.uv_required = uv_required

        self.expected_registration_client_extensions = expected_registration_client_extensions \
            if expected_registration_client_extensions \
            else DEFAULT_CLIENT_EXTENSIONS
        self.expected_registration_authenticator_extensions = expected_registration_authenticator_extensions \
            if expected_registration_authenticator_extensions \
            else DEFAULT_AUTHENTICATOR_EXTENSIONS

        if attestation_requirement_level not in ATTESTATION_REQUIREMENT_LEVELS:
            raise ValueError('Illegal attestation_requirement_level.')

        # With self attestation, the credential public key is also used as the attestation public key.
        self.self_attestation_permitted = attestation_requirement_level['self_attestation_permitted']
        self.trusted_attestation_cert_required = not self.self_attestation_permitted

        # With none attestation, authenticator attestation will not be performed.
        self.none_attestation_permitted = attestation_requirement_level['none_attestation_permitted']

    @staticmethod
    def parse_attestation_object(attestation_object):
        """
        Pull the individual fields out of an attestation object.

        :param attestation_object: The attestation object, as passed by the authenticator.
        :type attestation_object: basestring
        :return: The result of CBOR-decoding the attestation_object.
        :rtype: dict
        """

        return cbor2.loads(webauthn_b64_decode(attestation_object))

    @staticmethod
    def verify_attestation_statement(fmt, att_stmt, auth_data, client_data_hash=b'', none_attestation_permitted=True):
        """
        The procedure for verifying an attestation statement.

        The procedure returns either:
            * An error indicating that the attestation is invalid, or
            * The attestation type, and the trust path. This attestation trust path is
              either empty (in case of self attestation), an identifier of an ECDAA-Issuer
              public key (in the case of ECDAA), or a set of X.509 certificates.

        Verification of attestation objects requires that the Relying Party has a trusted
        method of determining acceptable trust anchors in step 15 above. Also, if
        certificates are being used, the Relying Party MUST have access to certificate
        status information for the intermediate CA certificates. The Relying Party MUST
        also be able to build the attestation certificate chain if the client did not
        provide this chain in the attestation information.

        In order to facilitate further checking of the certificate by prepolicies, the
        client_data_hash may be omitted, to only perform parsing of the statement, verifying
        it further down the line.

        :param fmt: The attestation format.
        :type fmt: basestring
        :param att_stmt: The attestation statement structure.
        :type att_stmt: dict
        :param auth_data: The authenticator data claimed to have been used for the attestation.
        :type auth_data: basestring
        :param client_data_hash: The hash of the serialized client data.
        :type client_data_hash: SHA256Type
        :param none_attestation_permitted: Whether to allow for the attestation type to be none or unsupported.
        :type none_attestation_permitted: bool
        :return: The attestation type, trust path, credential public key, credential id and aaguid.
        :rtype: (basestring, Certificate[], basestring, basestring, basestring)
        """

        attestation_data = auth_data[37:]
        aaguid = attestation_data[:16]
        credential_id_len = struct.unpack('!H', attestation_data[16:18])[0]
        cred_id = attestation_data[18:18 + credential_id_len]
        credential_pub_key = attestation_data[18 + credential_id_len:]

        if fmt == AttestationFormat.FIDO_U2F:
            # Step 1.
            #
            # Verify that attStmt is valid CBOR conforming to the syntax
            # defined above and perform CBOR decoding on it to extract the
            # contained fields.
            if 'x5c' not in att_stmt or 'sig' not in att_stmt:
                raise RegistrationRejectedException('Attestation statement must be a valid CBOR object.')

            # Step 2.
            #
            # Let attCert be the value of the first element of x5c. Let certificate
            # public key be the public key conveyed by attCert. If certificate public
            # key is not an Elliptic Curve (EC) public key over the P-256 curve,
            # terminate this algorithm and return an appropriate error.
            att_cert = att_stmt.get('x5c')[0]
            x509_att_cert = load_der_x509_certificate(att_cert, default_backend())
            certificate_public_key = x509_att_cert.public_key()
            if not isinstance(certificate_public_key.curve, SECP256R1):
                raise RegistrationRejectedException('Bad certificate public key.')

            # Step 3.
            #
            # Extract the claimed rpIdHash from authenticatorData, and the
            # claimed credentialId and credentialPublicKey from
            # authenticatorData.attestedCredentialData.
            #
            # The credential public key encoded in COSE_Key format, as defined in Section 7
            # of [RFC8152], using the CTAP2 canonical CBOR encoding form. The COSE_Key-encoded
            # credential public key MUST contain the optional "alg" parameter and MUST NOT
            # contain any other optional parameters. The "alg" parameter MUST contain a
            # COSEAlgorithmIdentifier value. The encoded credential public key MUST also
            # contain any additional required parameters stipulated by the relevant key type
            # specification, i.e., required for the key type "kty" and algorithm "alg" (see
            # Section 8 of [RFC8152]).
            try:
                public_key_alg, credential_public_key = _load_cose_public_key(credential_pub_key)
            except COSEKeyException as e:
                raise RegistrationRejectedException(str(e))

            public_key_u2f = _encode_public_key(credential_public_key)

            # Step 5.
            #
            # Let verificationData be the concatenation of (0x00 || rpIdHash ||
            # clientDataHash || credentialId || publicKeyU2F) (see Section 4.3
            # of [FIDO-U2F-Message-Formats]).
            auth_data_rp_id_hash = _get_auth_data_rp_id_hash(auth_data)
            alg = CoseAlgorithm.ES256
            signature = att_stmt['sig']
            verification_data = b''.join([
                b'\0',
                auth_data_rp_id_hash,
                client_data_hash,
                cred_id,
                public_key_u2f
            ])

            # Step 6.
            #
            # Verify the sig using verificationData and certificate public
            # key per [SEC1].
            # TODO: we need to check here what happens if we do not have a
            #  client_data_hash (this happens during the webauthntoken_allowed
            #  pre-policy).
            if client_data_hash:
                try:
                    _verify_signature(certificate_public_key, alg, verification_data, signature)
                except InvalidSignature:
                    raise RegistrationRejectedException('Invalid signature received.')
                except NotImplementedError:
                    # We do not support this. Treat as none attestation, if acceptable.
                    if none_attestation_permitted:
                        attestation_type = AttestationType.NONE
                        trust_path = []
                        return (
                            attestation_type,
                            trust_path,
                            credential_pub_key,
                            cred_id,
                            aaguid
                        )

                    raise RegistrationRejectedException('Unsupported algorithm.')

            # Step 7.
            #
            # If successful, return attestation type Basic with the
            # attestation trust path set to x5c.
            attestation_type = AttestationType.BASIC
            trust_path = [x509_att_cert]

            return (
                attestation_type,
                trust_path,
                credential_pub_key,
                cred_id,
                aaguid
            )
        elif fmt == AttestationFormat.PACKED:
            attestation_syntaxes = {
                AttestationType.BASIC: ([
                    'alg',
                    'x5c',
                    'sig'
                ]),
                AttestationType.ECDAA: ([
                    'alg',
                    'sig',
                    'ecdaaKeyId'
                ]),
                AttestationType.SELF_ATTESTATION: ([
                    'alg',
                    'sig'
                ])
            }

            # Step 1.
            #
            # Verify that attStmt is valid CBOR conforming to the syntax
            # defined above and perform CBOR decoding on it to extract the
            # contained fields.
            if set(att_stmt.keys()) not in [set(e) for e in attestation_syntaxes.values()]:
                raise RegistrationRejectedException('Attestation statement must be a valid CBOR object.')

            alg = att_stmt['alg']
            signature = att_stmt['sig']
            verification_data = b''.join([
                auth_data,
                client_data_hash
            ])

            # Step 2.
            #
            # If x5c is present, this indicates that the attestation
            # type is not ECDAA.
            if 'x5c' in att_stmt:
                # TODO: this could be a certificate chain, we should treat it as such
                att_cert = att_stmt['x5c'][0]
                x509_att_cert = load_der_x509_certificate(att_cert, default_backend())
                certificate_public_key = x509_att_cert.public_key()

                # Verify that sig is a valid signature over the
                # concatenation of authenticatorData and clientDataHash
                # using the attestation public key in attestnCert with
                # the algorithm specified in alg.
                #
                # client_data_hash will not be provided, if this function is
                # called just to parse out the attestation data, to check
                # against WEBAUTHNACTION.REQ. In that case, this function will
                # be called a second time later on for the actual validation of
                # the signature.
                if client_data_hash:
                    try:
                        _verify_signature(certificate_public_key, alg, verification_data, signature)
                    except InvalidSignature:
                        raise RegistrationRejectedException('Invalid signature received.')
                    except NotImplementedError:
                        # We do not support this. Treat as none attestation, if acceptable.
                        if none_attestation_permitted:
                            attestation_type = AttestationType.NONE
                            trust_path = []
                            return (
                                attestation_type,
                                trust_path,
                                credential_pub_key,
                                cred_id,
                                aaguid
                            )

                        raise RegistrationRejectedException('Unsupported algorithm.')

                #
                # Verify that attestnCert meets the requirements in
                # §8.2.1 Packed attestation statement certificate requirements.
                #
                # The attestation certificate MUST have the following
                # fields/extensions:
                #

                # Version MUST be set to 3 (which is indicated by an
                # ASN.1 INTEGER with value 2).
                if x509_att_cert.version.value != x509.Version.v3.value:
                    raise RegistrationRejectedException('Invalid attestation certificate version.')

                subject = x509_att_cert.subject
                c = subject.get_attributes_for_oid(x509.NameOID.COUNTRY_NAME)
                o = subject.get_attributes_for_oid(x509.NameOID.ORGANIZATION_NAME)
                ou = subject.get_attributes_for_oid(x509.NameOID.ORGANIZATIONAL_UNIT_NAME)
                cn = subject.get_attributes_for_oid(x509.NameOID.COMMON_NAME)

                # Subject-C: ISO 3166 code specifying the country
                #            where the Authenticator vendor is
                #            incorporated
                if not c:
                    raise RegistrationRejectedException('Attestation certificate must have subject-C.')

                # Subject-O: Legal name of the Authenticator vendor.
                if not o:
                    raise RegistrationRejectedException('Attestation certificate must have subject-O.')

                # Subject-OU: Literal string "Authenticator Attestation"
                if not ou or ou[0].value != 'Authenticator Attestation':
                    raise RegistrationRejectedException("Attestation certificate must have subject-OU set "
                                                        "to 'Authenticator Attestation'.")

                # Subject-CN: An UTF8String of the vendor's choosing.
                if not cn:
                    raise RegistrationRejectedException('Attestation certificate must have subject-CN.')

                extensions = x509_att_cert.extensions

                # If the related attestation root certificate is used
                # for multiple authenticator models, the Extension OID
                # 1.3.6.1.4.1.45724.1.1.4 (id-fido-gen-ce-aaguid) MUST
                # be present, containing the AAGUID as a 16-byte OCTET
                # STRING. The extension MUST NOT be marked as critical.
                try:
                    oid = x509.ObjectIdentifier('1.3.6.1.4.1.45724.1.1.4')
                    aaguid_ext = extensions.get_extension_for_oid(oid)
                    if aaguid_ext.value.value[2:] != aaguid:
                        raise RegistrationRejectedException('Attestation certificate AAGUID must match '
                                                            'authenticator data.')
                    if aaguid_ext.critical:
                        raise RegistrationRejectedException("Attestation certificate's 'id-fido-gen-ce-aaguid' "
                                                            "extension must not be marked critical.")
                except x509.ExtensionNotFound:
                    # This extension is optional.
                    pass

                # The Basic Constraints extension MUST have the CA
                # component set to false.
                bc_extension = extensions.get_extension_for_class(x509.BasicConstraints)
                if not bc_extension or bc_extension.value.ca:
                    raise RegistrationRejectedException('Attestation certificate must have Basic Constraints '
                                                        'extension with CA=false.')

                # If successful, return attestation type Basic and
                # attestation trust path x5c.
                attestation_type = AttestationType.BASIC
                trust_path = [x509_att_cert]

                return (
                    attestation_type,
                    trust_path,
                    credential_pub_key,
                    cred_id,
                    aaguid
                )
            elif 'ecdaaKeyId' in att_stmt:
                # We do not support this. If attestation is optional, have it go through anyway.
                if none_attestation_permitted:
                    attestation_type = AttestationType.ECDAA
                    trust_path = []
                    return (
                        attestation_type,
                        trust_path,
                        credential_pub_key,
                        cred_id,
                        aaguid
                    )

                # Step 3.
                #
                # If ecdaaKeyId is present, then the attestation type is
                # ECDAA. In this case:
                #   * Verify that sig is a valid signature over the
                #     concatenation of authenticatorData and clientDataHash
                #     using ECDAA-Verify with ECDAA-Issuer public key
                #     identified by ecdaaKeyId (see  [FIDOEcdaaAlgorithm]).
                #   * If successful, return attestation type ECDAA and
                #     attestation trust path ecdaaKeyId.
                raise RegistrationRejectedException('ECDAA attestation type is not currently supported.')
            else:
                # self-attestation
                # Step 1:
                # Validate that alg matches the algorithm of the credentialPublicKey
                # in authenticatorData.
                try:
                    public_key_alg, credential_public_key = _load_cose_public_key(credential_pub_key)
                except COSEKeyException as e:
                    raise RegistrationRejectedException(str(e))
                if alg != public_key_alg:
                    raise RegistrationRejectedException('credentialPublicKey algorithm {0!s} does '
                                                        'not match algorithm from attestation '
                                                        'statement {1!s}'.format(public_key_alg, alg))
                # Step 2:
                # Verify that sig is a valid signature over the concatenation of authenticatorData
                # and clientDataHash using the credential public key with alg.
                try:
                    _verify_signature(credential_public_key, alg, verification_data, signature)
                except InvalidSignature:
                    raise RegistrationRejectedException('Invalid signature received.')
                except NotImplementedError:  # pragma: no cover
                    log.warning('Unsupported algorithm ({0!s}) for signature '
                                'verification'.format(alg))
                    # We do not support this algorithm. Treat as none attestation, if acceptable.
                    if none_attestation_permitted:
                        return (
                            AttestationType.NONE,
                            [],
                            credential_pub_key,
                            cred_id,
                            aaguid
                        )
                    else:
                        raise RegistrationRejectedException('Unsupported algorithm '
                                                            '({0!s}).'.format(alg))
                return (
                    AttestationType.SELF_ATTESTATION,
                    [],
                    credential_pub_key,
                    cred_id,
                    aaguid
                )
        else:
            # Attestation is either none, or unsupported.
            if not none_attestation_permitted:
                if fmt == AttestationFormat.NONE:
                    raise RegistrationRejectedException('Authenticator attestation is required.')
                else:
                    raise RegistrationRejectedException(
                        'Unsupported authenticator attestation format ({0!s})!'.format(fmt))

            # Treat as none attestation.
            #
            # Step 1.
            #
            # Return attestation type None with an empty trust path.
            attestation_type = AttestationType.NONE
            trust_path = []
            return (
                attestation_type,
                trust_path,
                credential_pub_key,
                cred_id,
                aaguid
            )

    def verify(self, existing_credential_ids=None):
        """
        Verify the WebAuthnRegistrationResponse.

        This will perform the registration ceremony for the
        WebAuthnRegistrationResponse. It will only return on successful
        verification. In any other case, an appropriate error will be raised.

        :param existing_credential_ids: A list of existing credential ids to check for duplicates.
        :type existing_credential_ids: basestring[]
        :return: The WebAuthnCredential produced by the registration ceremony.
        :rtype: WebAuthnCredential
        """

        try:
            # As described in https://www.w3.org/TR/webauthn/#sctn-registering-a-new-credential
            # In the docs it starts at step 5.
            #
            # Step 1.
            #
            # Let JSONtext be the result of running UTF-8 decode on the value of
            # response.clientDataJSON
            json_text = self.registration_response.get('clientData', '')

            # Step 2.
            #
            # Let C, the client data claimed as collected during the credential
            # creation, be the result of running an implementation-specific JSON
            # parser on JSONtext.
            decoded_cd = webauthn_b64_decode(json_text)
            c = json.loads(to_unicode(decoded_cd))

            # Step 3.
            #
            # Verify that the value of C.type is webauthn.create.
            if not _verify_type(c.get('type'), ClientDataType.CREATE):
                raise RegistrationRejectedException('Invalid type.')

            # Step 4.
            #
            # Verify that the value of C.challenge matches the challenge that was sent
            # to the authenticator in the create() call.
            if not _verify_challenge(c.get('challenge'), self.challenge):
                raise RegistrationRejectedException('Unable to verify challenge.')

            # Step 5.
            #
            # Verify that the value of C.origin matches the Relying Party's origin.
            if not _verify_origin(c, self.origin):
                raise RegistrationRejectedException('Unable to verify origin.')

            # Step 6.
            #
            # Verify that the value of C.tokenBinding.status matches the state of
            # Token Binding for the TLS connection over which the assertion was
            # obtained. If Token Binding was used on that TLS connection, also verify
            # that C.tokenBinding.id matches the base64url encoding of the Token
            # Binding ID for the connection.

            # Chrome does not currently supply token binding in the clientDataJSON
            # if not _verify_token_binding_id(c):
            #     raise RegistrationRejectedException('Unable to verify token binding ID.')

            # Step 7.
            #
            # Compute the hash of response.clientDataJSON using SHA-256.
            client_data_hash = _get_client_data_hash(decoded_cd)

            # Step 8.
            #
            # Perform CBOR decoding on the attestationObject field of
            # the AuthenticatorAttestationResponse structure to obtain
            # the attestation statement format fmt, the authenticator
            # data authData, and the attestation statement attStmt.
            att_obj = self.parse_attestation_object(self.registration_response.get('attObj'))
            att_stmt = att_obj.get('attStmt')
            auth_data = att_obj.get('authData')
            fmt = att_obj.get('fmt')
            if not auth_data or len(auth_data) < 37:
                raise RegistrationRejectedException('Auth data must be at least 37 bytes.')

            # Step 9.
            #
            # Verify that the RP ID hash in authData is indeed the
            # SHA-256 hash of the RP ID expected by the RP.
            if not _verify_rp_id_hash(_get_auth_data_rp_id_hash(auth_data), self.rp_id):
                raise RegistrationRejectedException('Unable to verify RP ID hash.')

            # Step 10.
            #
            # Verify that the User Present bit of the flags in authData
            # is set.
            if not AuthenticatorDataFlags(auth_data).user_present:
                raise RegistrationRejectedException('Malformed request received.')

            # Step 11.
            #
            # If user verification is required for this registration, verify
            # that the User Verified bit of the flags in authData is set.
            if self.uv_required and not AuthenticatorDataFlags(auth_data).user_verified:
                raise RegistrationRejectedException('Malformed request received.')

            # Step 12.
            #
            # Verify that the values of the client extension outputs in
            # clientExtensionResults and the authenticator extension outputs
            # in the extensions in authData are as expected, considering the
            # client extension input values that were given as the extensions
            # option in the create() call. In particular, any extension
            # identifier values in the clientExtensionResults and the extensions
            # in authData MUST be also be present as extension identifier values
            # in the extensions member of options, i.e., no extensions are
            # present that were not requested. In the general case, the meaning
            # of "are as expected" is specific to the Relying Party and which
            # extensions are in use.
            if not _verify_authenticator_extensions(auth_data, self.expected_registration_authenticator_extensions):
                raise RegistrationRejectedException('Unable to verify authenticator extensions.')
            if not _verify_client_extensions(
                    self.registration_response.get('registrationClientExtensions'),
                    self.expected_registration_client_extensions
            ):
                raise RegistrationRejectedException('Unable to verify client extensions.')

            # Step 13.
            #
            # Determine the attestation statement format by performing
            # a USASCII case-sensitive match on fmt against the set of
            # supported WebAuthn Attestation Statement Format Identifier
            # values. The up-to-date list of registered WebAuthn
            # Attestation Statement Format Identifier values is maintained
            # in the IANA registry of the same name.
            if not _verify_attestation_statement_format(fmt):
                raise RegistrationRejectedException('Unable to verify attestation statement format.')

            # Step 14.
            #
            # Verify that attStmt is a correct attestation statement, conveying
            # a valid attestation signature, by using the attestation statement
            # format fmt's verification procedure given attStmt, authData and
            # the hash of the serialized client data computed in step 7.
            (
                attestation_type,
                trust_path,
                credential_public_key,
                cred_id,
                aaguid
            ) = self.verify_attestation_statement(
                fmt,
                att_stmt,
                auth_data,
                client_data_hash,
                self.none_attestation_permitted
            )
            b64_cred_id = webauthn_b64_encode(cred_id)

            # Step 15.
            #
            # If validation is successful, obtain a list of acceptable trust
            # anchors (attestation root certificates or ECDAA-Issuer public
            # keys) for that attestation type and attestation statement format
            # fmt, from a trusted source or from policy. For example, the FIDO
            # Metadata Service [FIDOMetadataService] provides one way to obtain
            # such information, using the aaguid in the attestedCredentialData
            # in authData.
            trust_anchors = _get_trust_anchors(attestation_type, fmt, self.trust_anchor_dir) \
                if self.trust_anchor_dir \
                else None
            if not trust_anchors and self.trusted_attestation_cert_required:
                raise RegistrationRejectedException('No trust anchors available to verify attestation certificate.')

            # Step 16.
            #
            # Assess the attestation trustworthiness using the outputs of the
            # verification procedure in step 14, as follows:
            #
            #     * If self attestation was used, check if self attestation is
            #       acceptable under Relying Party policy.
            #     * If ECDAA was used, verify that the identifier of the
            #       ECDAA-Issuer public key used is included in the set of
            #       acceptable trust anchors obtained in step 15.
            #     * Otherwise, use the X.509 certificates returned by the
            #       verification procedure to verify that the attestation
            #       public key correctly chains up to an acceptable root
            #       certificate.
            if attestation_type == AttestationType.SELF_ATTESTATION and not self.self_attestation_permitted:
                raise RegistrationRejectedException('Self attestation is not permitted.')
            is_trusted_attestation_cert = ((attestation_type == AttestationType.BASIC
                                            and _is_trusted_x509_attestation_cert(trust_path, trust_anchors))
                                           or (attestation_type == AttestationType.ECDAA
                                               and _is_trusted_ecdaa_attestation_certificate(None, trust_anchors)))
            is_signed_attestation_cert = attestation_type in SUPPORTED_ATTESTATION_TYPES

            if is_trusted_attestation_cert:
                attestation_level = AttestationLevel.TRUSTED
            elif is_signed_attestation_cert:
                attestation_level = AttestationLevel.UNTRUSTED
            else:
                attestation_level = AttestationLevel.NONE

            # Step 17.
            #
            # Check that the credentialId is not yet registered to any other user.
            # If registration is requested for a credential that is already registered
            # to a different user, the Relying Party SHOULD fail this registration
            # ceremony, or it MAY decide to accept the registration, e.g. while deleting
            # the older registration.
            if existing_credential_ids and b64_cred_id in existing_credential_ids:
                raise RegistrationRejectedException('Credential already exists.')

            # Step 18.
            #
            # If the attestation statement attStmt verified successfully and is
            # found to be trustworthy, then register the new credential with the
            # account that was denoted in the options.user passed to create(),
            # by associating it with the credentialId and credentialPublicKey in
            # the attestedCredentialData in authData, as appropriate for the
            # Relying Party's system.
            credential = WebAuthnCredential(rp_id=self.rp_id,
                                            origin=self.origin,
                                            aaguid=aaguid,
                                            credential_id=b64_cred_id,
                                            public_key=webauthn_b64_encode(credential_public_key),
                                            sign_count=struct.unpack('!I', auth_data[33:37])[0],
                                            attestation_level=attestation_level,
                                            attestation_cert=trust_path[0] if trust_path else None)
            if is_trusted_attestation_cert:
                return credential

            # Step 19.
            #
            # If the attestation statement attStmt successfully verified but is
            # not trustworthy per step 16 above, the Relying Party SHOULD fail
            # the registration ceremony.
            #
            #     NOTE: However, if permitted by policy, the Relying Party MAY
            #           register the credential ID and credential public key but
            #           treat the credential as one with self attestation (see
            #           6.3.3 Attestation Types). If doing so, the Relying Party
            #           is asserting there is no cryptographic proof that the
            #           public key credential has been generated by a particular
            #           authenticator model. See [FIDOSecRef] and [UAFProtocol]
            #           for a more detailed discussion.
            if self.trusted_attestation_cert_required:
                raise RegistrationRejectedException('Untrusted attestation certificate.')
            if not is_signed_attestation_cert and not self.none_attestation_permitted:
                raise RegistrationRejectedException('No (or unsupported) attestation certificate.')
            return credential

        except Exception as e:
            raise RegistrationRejectedException('Registration rejected. Error: {}'.format(e))


class WebAuthnAssertionResponse:
    """
    The WebAuthn assertion response containing all information needed to verify the authentication ceremony.
    """

    def __init__(self,
                 webauthn_user,
                 assertion_response,
                 challenge,
                 origin,
                 allow_credentials=None,
                 uv_required=False,
                 expected_assertion_client_extensions=None,
                 expected_assertion_authenticator_extensions=None):
        """
        Create a new WebAuthnAssertionResponse object.

        :param webauthn_user: The WebAuthnUser used to create the assertion.
        :type webauthn_user: WebAuthnUser
        :param assertion_response: The assertion as a public key credential dictionary.
        :type assertion_response: dict
        :param challenge: The challenge that was sent to the client.
        :type challenge: basestring
        :param origin: The origin of the user.
        :type origin: basestring
        :param allow_credentials: Which existing credentials to allow for the authentication.
        :type allow_credentials: list of basestring
        :param uv_required: Whether user verification is required, preferred or discouraged.
        :type uv_required: UserVerificationLevel
        :param expected_assertion_client_extensions: A dict whose keys indicate which client extensions are expected.
        :type expected_assertion_client_extensions: dict
        :param expected_assertion_authenticator_extensions: A dict whose keys indicate which auth exts to expect.
        :type expected_assertion_authenticator_extensions: dict
        :return: A WebAuthnAssertionResponse object.
        :rtype: WebAuthnAssertionResponse
        """
        self.assertion_response = assertion_response
        self.challenge = challenge
        self.origin = origin
        self.allow_credentials = allow_credentials
        self.uv_required = uv_required or UserVerificationLevel.PREFERRED

        self.expected_assertion_client_extensions = expected_assertion_client_extensions or DEFAULT_CLIENT_EXTENSIONS
        self.expected_assertion_authenticator_extensions = (expected_assertion_authenticator_extensions
                                                            or DEFAULT_AUTHENTICATOR_EXTENSIONS)

        if not isinstance(webauthn_user, WebAuthnUser):
            raise ValueError('Invalid user type.')
        self.webauthn_user = webauthn_user

    def verify(self):
        """
        Verify the WebAuthnAssertionResponse.

        This will perform the authentication ceremony for the
        WebAuthnAssertionResponse. It will only return on successful
        verification. In any other case, an appropriate error will be raised.

        :return: The new sign count of the authenticated credential.
        :rtype: int
        """
        try:
            # Step 1.
            #
            # If the allowCredentials option was given when this authentication
            # ceremony was initiated, verify that credential.id identifies one
            # of the public key credentials that were listed in allowCredentials.
            if self.allow_credentials and self.assertion_response.get('id') not in self.allow_credentials:
                raise AuthenticationRejectedException('Credential not allowed.')

            # Step 2.
            #
            # If credential.response.userHandle is present, verify that the user
            # identified by this value is the owner of the public key credential
            # identified by credential.id.
            user_handle = self.assertion_response.get('userHandle')
            try:
                user_handle = base64url_to_bytes(user_handle).decode("utf-8") if user_handle else None
            except (ValueError, TypeError) as e:
                raise AuthenticationRejectedException(f"Malformed user handle: {e}")
            if user_handle and not user_handle == self.webauthn_user.user_id:
                raise AuthenticationRejectedException(f"Invalid credential: user handle from param {user_handle} "
                                                      f"does not match the one from the user "
                                                      f"{self.webauthn_user.user_id}.")

            # Step 3.
            #
            # Using credential's id attribute (or the corresponding rawId, if
            # base64url encoding is inappropriate for your use case), look up
            # the corresponding credential public key.
            if not _validate_credential_id(self.webauthn_user.credential_id):
                raise AuthenticationRejectedException('Invalid credential ID.')
            if not self.webauthn_user.public_key:
                raise WebAuthnUserDataMissing("public_key missing.")
            public_key_alg, user_pubkey = _load_cose_public_key(webauthn_b64_decode(self.webauthn_user.public_key))

            # Step 4.
            #
            # Let cData, aData and sig denote the value of credential's
            # response's clientDataJSON, authenticatorData, and signature
            # respectively.
            c_data = webauthn_b64_decode(self.assertion_response.get('clientData'))
            a_data = webauthn_b64_decode(self.assertion_response.get('authData'))
            sig = webauthn_b64_decode(self.assertion_response.get('signature'))

            # Step 5.
            #
            # Let JSONtext be the result of running UTF-8 decode on the
            # value of cData.
            json_text = to_unicode(c_data)

            # Step 6.
            #
            # Let C, the client data claimed as used for the signature,
            # be the result of running an implementation-specific JSON
            # parser on JSONtext.
            c = json.loads(json_text)

            # Step 7.
            #
            # Verify that the value of C.type is the string webauthn.get.
            if not _verify_type(c.get('type'), ClientDataType.GET):
                raise RegistrationRejectedException('Invalid type.')

            # Step 8.
            #
            # Verify that the value of C.challenge matches the challenge
            # that was sent to the authenticator in the
            # PublicKeyCredentialRequestOptions passed to the get() call.
            if not _verify_challenge(c.get('challenge'), self.challenge):
                raise AuthenticationRejectedException('Unable to verify challenge.')

            # Step 9.
            #
            # Verify that the value of C.origin matches the Relying
            # Party's origin.
            if not _verify_origin(c, self.origin):
                raise AuthenticationRejectedException('Unable to verify origin.')

            # Step 10.
            #
            # Verify that the value of C.tokenBinding.status matches
            # the state of Token Binding for the TLS connection over
            # which the attestation was obtained. If Token Binding was
            # used on that TLS connection, also verify that
            # C.tokenBinding.id matches the base64url encoding of the
            # Token Binding ID for the connection.

            # XXX: Chrome does not currently supply token binding in the clientDataJSON
            # if not _verify_token_binding_id(c):
            #     raise AuthenticationRejectedException('Unable to verify token binding ID.')

            # Step 11.
            #
            # Verify that the rpIdHash in aData is the SHA-256 hash of
            # the RP ID expected by the Relying Party.
            if not _verify_rp_id_hash(_get_auth_data_rp_id_hash(a_data), self.webauthn_user.rp_id):
                raise AuthenticationRejectedException('Unable to verify RP ID hash.')

            # Step 12.
            #
            # Verify that the User Present bit of the flags in authData
            # is set.
            if not AuthenticatorDataFlags(a_data).user_present:
                raise AuthenticationRejectedException('Malformed request received.')

            # Step 13.
            #
            # If user verification is required for this assertion, verify that
            # the User Verified bit of the flags in authData is set.
            if (self.uv_required == UserVerificationLevel.REQUIRED
                    and not AuthenticatorDataFlags(a_data).user_verified):
                raise RegistrationRejectedException('Malformed request received.')

            # Step 14.
            #
            # Verify that the values of the client extension outputs in
            # clientExtensionResults and the authenticator extension outputs
            # in the extensions in authData are as expected, considering the
            # client extension input values that were given as the extensions
            # option in the get() call. In particular, any extension identifier
            # values in the clientExtensionResults and the extensions in
            # authData MUST be also be present as extension identifier values
            # in the extensions member of options, i.e., no extensions are
            # present that were not requested. In the general case, the meaning
            # of "are as expected" is specific to the Relying Party and which
            # extensions are in use.
            if not _verify_authenticator_extensions(a_data, self.expected_assertion_authenticator_extensions):
                raise AuthenticationRejectedException('Unable to verify authenticator extensions.')
            if not _verify_client_extensions(
                    self.assertion_response.get('assertionClientExtensions'),
                    self.expected_assertion_client_extensions
            ):
                raise AuthenticationRejectedException('Unable to verify client extensions.')

            # Step 15.
            #
            # Let hash be the result of computing a hash over the cData
            # using SHA-256.
            client_data_hash = _get_client_data_hash(c_data)

            # Step 16.
            #
            # Using the credential public key looked up in step 3, verify
            # that sig is a valid signature over the binary concatenation
            # of aData and hash.
            try:
                _verify_signature(public_key=user_pubkey,
                                  alg=public_key_alg,
                                  data=b''.join([
                                      a_data,
                                      client_data_hash
                                  ]),
                                  signature=sig)
            except InvalidSignature:
                raise AuthenticationRejectedException('Invalid signature received.')
            except NotImplementedError:
                raise AuthenticationRejectedException('Unsupported algorithm.')

            # Step 17.
            #
            # If the signature counter value adata.signCount is nonzero or
            # the value stored in conjunction with credential's id attribute
            # is nonzero, then run the following sub-step:
            #     If the signature counter value adata.signCount is
            #         greater than the signature counter value stored in
            #         conjunction with credential's id attribute.
            #             Update the stored signature counter value,
            #             associated with credential's id attribute,
            #             to be the value of adata.signCount.
            #         less than or equal to the signature counter value
            #         stored in conjunction with credential's id attribute.
            #             This is a signal that the authenticator may be
            #             cloned, i.e. at least two copies of the credential
            #             private key may exist and are being used in parallel.
            #             Relying Parties should incorporate this information
            #             into their risk scoring. Whether the Relying Party
            #             updates the stored signature counter value in this
            #             case, or not, or fails the authentication ceremony
            #             or not, is Relying Party-specific.
            sign_count = struct.unpack('!I', a_data[33:37])[0]
            if (sign_count != 0 or self.webauthn_user.sign_count != 0) and sign_count <= self.webauthn_user.sign_count:
                raise AuthenticationRejectedException('Duplicate authentication detected.')

            # Step 18.
            #
            # If all the above steps are successful, continue with the
            # authentication ceremony as appropriate. Otherwise, fail the
            # authentication ceremony.
            return sign_count

        except Exception as e:
            raise AuthenticationRejectedException('Authentication rejected. Error: {}'.format(e))


def webauthn_b64_decode(encoded):
    """
    Pad a WebAuthn base64-encoded string and decode it.

    WebAuthn specifies a web-safe base64 encoding *without* padding. Since
    this is the same as u2f, this function will just rely on the existing u2f
    implementation of this algorithm.

    :param encoded: A WebAuthn base64-encoded string.
    :type encoded: basestring or bytes
    :return: The decoded binary.
    :rtype: bytes
    """

    return url_decode(encoded)


def webauthn_b64_encode(raw):
    """
    Encode bytes using WebAuthn base64-encoding.

    WebAuthn specifies a web-safe base64 encoding *without* padding. Since
    this is the same as u2f, this function will just rely on the existing u2f
    implementation of this algorithm.

    :param raw: Bytes to encode.
    :type raw: basestring or bytes
    :return: The encoded base64.
    :rtype: basestring
    """

    return url_encode(raw)


def _encode_public_key(public_key):
    """
    Extracts the x and y coordinates from a public point on a Cryptography elliptic curve.

    The result of running this function is a 65 byte string. This function is the inverse of
    decode_public_key().public_key().

    :param public_key: An EllipticCurvePublicKey object
    :return: The coordinates packed into a standard byte string representation.
    """
    numbers = public_key.public_numbers()
    return b'\x04' + binascii.unhexlify('{:064x}{:064x}'.format(numbers.x, numbers.y))


def _load_cose_public_key(key_bytes):
    cose_public_key = cbor2.loads(key_bytes)

    if CosePublicKey.ALG not in cose_public_key:
        raise COSEKeyException('Public key missing required algorithm parameter.')

    alg = cose_public_key[CosePublicKey.ALG]

    if alg == CoseAlgorithm.ES256:

        required_keys = {
            CosePublicKey.ALG,
            CosePublicKey.X,
            CosePublicKey.Y
        }

        if not set(cose_public_key.keys()).issuperset(required_keys):
            raise COSEKeyException('Public key must match COSE_Key spec.')

        if len(cose_public_key[CosePublicKey.X]) != 32:
            raise RegistrationRejectedException('Bad public key.')
        x = int(codecs.encode(cose_public_key[CosePublicKey.X], 'hex'), 16)

        if len(cose_public_key[CosePublicKey.Y]) != 32:
            raise RegistrationRejectedException('Bad public key.')
        y = int(codecs.encode(cose_public_key[CosePublicKey.Y], 'hex'), 16)

        return alg, EllipticCurvePublicNumbers(x, y, SECP256R1()).public_key(backend=default_backend())
    elif alg in (CoseAlgorithm.PS256, CoseAlgorithm.RS256, CoseAlgorithm.RS1):

        required_keys = {
            CosePublicKey.ALG,
            CosePublicKey.E,
            CosePublicKey.N
        }

        if not set(cose_public_key.keys()).issuperset(required_keys):
            raise COSEKeyException('Public key must match COSE_Key spec.')

        if len(cose_public_key[CosePublicKey.E]) != 3 or len(cose_public_key[CosePublicKey.N]) != 256:
            raise COSEKeyException('Bad public key.')

        e = int(codecs.encode(cose_public_key[CosePublicKey.E], 'hex'), 16)
        n = int(codecs.encode(cose_public_key[CosePublicKey.N], 'hex'), 16)

        return alg, RSAPublicNumbers(e, n).public_key(backend=default_backend())
    else:
        log.warning('Unsupported webAuthn COSE algorithm: {0!s}'.format(alg))
        raise COSEKeyException('Unsupported algorithm.')


def _get_trust_anchors(attestation_type, attestation_fmt, trust_anchor_dir):
    """
    Return a list of trusted attestation root certificates.

    This will fetch all CA certificates from the given directory, silently
    skipping any invalid ones.

    :param attestation_type: The attestation type being used. If the type is
                             unsupported, an empty list is returned.
    :type attestation_type: str
    :param attestation_fmt: The attestation format being used. If the format is
                            unsupported, an empty list is returned.
    :type attestation_fmt: str
    :param trust_anchor_dir: The path to the directory that contains the CA certificates.
    :type trust_anchor_dir: str
    :return: The list of trust anchors.
    :rtype: list
    """

    if attestation_type not in SUPPORTED_ATTESTATION_TYPES \
            or attestation_fmt not in SUPPORTED_ATTESTATION_FORMATS:
        log.debug('Unsupported attestation type ({0!s}) or attestation format '
                  '({1!s}).'.format(attestation_type, attestation_fmt))
        return []

    trust_anchors = []

    if os.path.isdir(trust_anchor_dir):
        for trust_anchor_name in os.listdir(trust_anchor_dir):
            trust_anchor_path = os.path.join(trust_anchor_dir, trust_anchor_name)
            if os.path.isfile(trust_anchor_path):
                try:
                    with open(trust_anchor_path, 'rb') as f:
                        pem_data = f.read().strip()
                        pem = cryptography.x509.load_pem_x509_certificate(pem_data.strip())
                        trust_anchors.append(pem)
                except Exception as e:
                    log.info('Could not load certificate {0!s}: '
                             '{1!s}'.format(trust_anchor_path, e))
    else:
        log.debug('Trust anchor directory ({0!s}) not available.'.format(trust_anchor_dir))

    return trust_anchors


def _is_trusted_x509_attestation_cert(trust_path, trust_anchors):
    if not trust_path or not isinstance(trust_path, list) or not trust_anchors or not isinstance(trust_anchors, list):
        return False

    # TODO: this could be a certificate chain. We should treat it as such
    attestation_cert = trust_path[0]
    store = crypto.X509Store()
    # Since the certificates are in pyca.cryptography format, we need to convert
    # them to the OpenSSL.crypto format
    for i in trust_anchors:
        store.add_cert(crypto.X509.from_cryptography(i))
    store_ctx = crypto.X509StoreContext(store, crypto.X509.from_cryptography(attestation_cert))

    try:
        store_ctx.verify_certificate()
        return True
    except Exception as e:
        log.info('Unable to verify certificate: {}'.format(e))

    return False


def _is_trusted_ecdaa_attestation_certificate(ecdaa_issuer_public_key, trust_anchors):
    # TODO: implement
    raise NotImplementedError


def _verify_type(received_type, expected_type):
    return received_type == expected_type


def _verify_challenge(received_challenge, sent_challenge):
    return received_challenge \
        and sent_challenge \
        and isinstance(received_challenge, str) \
        and isinstance(sent_challenge, str) \
        and constant_time.bytes_eq(
            to_bytes(sent_challenge),
            to_bytes(received_challenge)
        )


def _verify_origin(client_data, origin):
    return isinstance(client_data, dict) \
        and client_data.get('origin') \
        and client_data.get('origin') == origin


def _verify_token_binding_id(client_data):
    """
    Verify tokenBinding. Currently, this is unimplemented, so it will simply
    return false if tokenBinding is required.

    The tokenBinding member contains information about the state of the
    Token Binding protocol used when communicating with the Relying Party.
    The status member is one of:
        not-supported: when the client does not support token binding.
            supported: the client supports token binding, but it was not
                       negotiated when communicating with the Relying
                       Party.
              present: token binding was used when communicating with the
                       Relying Party. In this case, the id member MUST be
                       present and MUST be a base64url encoding of the
                       Token Binding ID that was used.

    :param client_data: The WebAuthn client data dictionary.
    :type client_data: dict
    :return: False, if tokenBinding is present.
    :rtype: bool
    """

    return client_data['tokenBinding']['status'] in ('supported', 'not_supported')


def _verify_client_extensions(client_extensions, expected_client_extensions):
    """
    Verify the client extensions.

    This will verify that no additional extensions were provided, that were not requested.  The extensions will be
    passed in as provided by the authenticator. Any parsing is done inside the function.

    :param client_extensions: The registrationClientExtensions or assertionClientExtensions field, respectively.
    :type client_extensions: basestring
    :param expected_client_extensions: A dictionary whose keys indicate the extensions to expect.
    :type expected_client_extensions: dict
    :return: Whether there were any unexpected extensions.
    :rtype: bool
    """
    return not client_extensions \
        or set(expected_client_extensions.keys()).issuperset(json.loads(client_extensions).keys())


def _verify_authenticator_extensions(auth_data, expected_authenticator_extensions):
    """
    Verify the authenticator extensions.

    This implementation does not currently support any authenticator
    extensions, so the authenticator is expected to never send any. Thus,
    this function will simply return false if there is any authenticator
    extensions at all for now.

    :param auth_data: The authenticator data.
    :type auth_data: bytes
    :param expected_authenticator_extensions: A dictionary whose keys indicate the extensions to expect.
    :type expected_authenticator_extensions: dict
    :return: Whether there were any unexpected extensions.
    :rtype: bool
    """
    return not AuthenticatorDataFlags(auth_data).extension_data_included


def _verify_rp_id_hash(auth_data_rp_id_hash, rp_id):
    rp_id_hash = hashlib.sha256(to_bytes(rp_id)).digest()
    return constant_time.bytes_eq(auth_data_rp_id_hash, rp_id_hash)


def _verify_attestation_statement_format(fmt):
    """
    Verify that the attestation statement format identifier matches a registered attestation statement format.

    :param fmt: The attestation statement format.
    :type fmt: basestring
    :return: Whether the attestation statement format is supported.
    :rtype: bool
    """

    return isinstance(fmt, str) and fmt in REGISTERED_ATTESTATION_FORMATS


def _get_auth_data_rp_id_hash(auth_data):
    if not isinstance(auth_data, bytes):
        return False

    return auth_data[:32]


def _get_client_data_hash(decoded_client_data):
    """
    Compute the SHA256 hash of the client data.

    :param decoded_client_data: The client data to hash.
    :type decoded_client_data: bytes
    :return: The hash of the client data.
    :rtype: bytes
    """
    if not isinstance(decoded_client_data, bytes):
        return ''

    return hashlib.sha256(decoded_client_data).digest()


def _validate_credential_id(credential_id):
    return isinstance(credential_id, str)


def _verify_signature(public_key, alg, data, signature):
    if alg == CoseAlgorithm.ES256:
        public_key.verify(signature, data, ECDSA(SHA256()))
    elif alg == CoseAlgorithm.RS256:
        public_key.verify(signature, data, PKCS1v15(), SHA256())
    elif alg == CoseAlgorithm.PS256:
        padding = PSS(mgf=MGF1(SHA256()), salt_length=PSS.MAX_LENGTH)
        public_key.verify(signature, data, padding, SHA256())
    elif alg == CoseAlgorithm.RS1:
        public_key.verify(signature, data, PKCS1v15(), SHA1())  # nosec B303 # part of webauthn specification
    else:
        raise NotImplementedError()



# SPDX-FileCopyrightText: (C) 2023 Paul Lettich <paul.lettich@netknights.it>
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Info: https://privacyidea.org
#
# This code is free software: you can redistribute it and/or
# modify it under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation, either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program. If not, see <http://www.gnu.org/licenses/>.

from flask.cli import AppGroup
import click
from privacyidea.lib.auth import (create_db_admin, list_db_admin,
                                  delete_db_admin)

admin_cli = AppGroup("admin", help="Manage local administrators")


@admin_cli.command("add")
@click.argument('username')
@click.option("-e", "--email", "email")
@click.password_option('-p', '--password')
def add_admin(username, email, password):
    """
    Register a new administrator in the database.
    """
    create_db_admin(username, email, password)
    click.echo('Admin {0} was registered successfully.'.format(username))


@admin_cli.command("list")
def list_admin():
    """
    List all administrators.
    """
    admins = list_db_admin()
    click.echo("Name \t email")
    click.echo(30 * "=")
    for admin in admins:
        click.echo(f"{admin.username} \t {admin.email}")


@admin_cli.command("delete")
@click.argument("username")
def delete_admin(username):
    """
    Delete an existing administrator.
    """
    delete_db_admin(username)


@admin_cli.command("change")
@click.argument('username')
@click.option("-e", "--email", "email")
@click.password_option()
def change_admin(username, email, password):
    """
    Change the email address or the password of an existing administrator USERNAME.
    """
    create_db_admin(username, email, password)



"""Add column used_login to usercache

Revision ID: cb6d7b7bae63
Revises: a63df077051a
Create Date: 2018-12-18 11:59:27.486883

"""

# revision identifiers, used by Alembic.
revision = 'cb6d7b7bae63'
down_revision = 'a63df077051a'

from alembic import op
import sqlalchemy as sa


def upgrade():
    try:
        op.add_column('usercache', sa.Column('used_login', sa.Unicode(length=64), nullable=True))
        op.create_index(op.f('ix_usercache_used_login'), 'usercache', ['used_login'], unique=False)
    except Exception as exx:
        print('Adding of column "used_login" in table usercache failed: {!r}'.format(exx))
        print('This is expected behavior if this column already exists.')


def downgrade():
    op.drop_index(op.f('ix_usercache_used_login'), table_name='usercache')
    op.drop_column('usercache', 'used_login')




# (c) NetKnights GmbH 2024,  https://netknights.it
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-FileCopyrightText: 2024 Nils Behlen <nils.behlen@netknights.it>
# SPDX-FileCopyrightText: 2024 Jelina Unger <jelina.unger@netknights.it>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
import datetime
import importlib
import json
import logging
import os
from typing import Union

from sqlalchemy import func
from sqlalchemy.orm import Query

from privacyidea.api.lib.utils import send_result
from privacyidea.lib.challenge import delete_challenges
from privacyidea.lib.config import get_from_config
from privacyidea.lib.containerclass import TokenContainerClass
from privacyidea.lib.containers.container_info import PI_INTERNAL, TokenContainerInfoData
from privacyidea.lib.containertemplate.containertemplatebase import ContainerTemplateBase
from privacyidea.lib.error import ResourceNotFoundError, ParameterError, EnrollmentError, UserError, PolicyError
from privacyidea.lib.log import log_with
from privacyidea.lib.machine import is_offline_token
from privacyidea.lib.token import (get_tokens_from_serial_or_user, get_tokens,
                                   convert_token_objects_to_dicts, init_token)
from privacyidea.lib.user import User
from privacyidea.lib.utils import hexlify_and_unicode
from privacyidea.models import (TokenContainer, TokenContainerOwner, Token, TokenContainerToken,
                                Realm, TokenContainerTemplate)

log = logging.getLogger(__name__)


def delete_container_by_id(container_id: int) -> int:
    """
    Delete the container with the given id. If it does not exist, raises a ResourceNotFoundError.

    :param container_id: The id of the container to delete
    :return: ID of the deleted container on success
    """
    if not container_id:
        raise ParameterError("Unable to delete container without id.")

    container = find_container_by_id(container_id)

    # Delete challenges
    delete_challenges(serial=container.serial)

    return container.delete()


def delete_container_by_serial(serial: str) -> int:
    """
    Delete the container with the given serial. If it does not exist, raises a ResourceNotFoundError.

    :param serial: The serial of the container to delete
    :return: ID of the deleted container on success
    """
    if not serial:
        raise ParameterError("Unable to delete container without serial.")
    container = find_container_by_serial(serial)

    # Delete challenges
    delete_challenges(serial=serial)

    return container.delete()


def _gen_serial(container_type: str) -> str:
    """
    Generate a new serial for a container of the given type

    :param container_type: The type of the container
    :return: The generated serial
    """
    serial_len = int(get_from_config("SerialLength") or 8)
    prefix = "CONT"
    for ctype, cls in get_container_classes().items():
        if ctype.lower() == container_type.lower():
            prefix = cls.get_class_prefix()

    container_num = TokenContainer.query.filter(TokenContainer.type == container_type).count()
    while True:
        rnd = ""
        count = '{:04d}'.format(container_num)
        rnd_len = serial_len - len(count)
        if rnd_len > 0:
            rnd = hexlify_and_unicode(os.urandom(rnd_len)).upper()[0:rnd_len]
        serial = f"{prefix}{count}{rnd}"
        if not TokenContainer.query.filter(TokenContainer.serial == serial).first():
            break

    return serial


def create_container_from_db_object(db_container: TokenContainer) -> Union[TokenContainerClass, None]:
    """
    Create a TokenContainerClass object from the given db object.

    :param db_container: The db object to create the container from
    :return: The created container object or None if the container type is not supported
    """
    for ctypes, cls in get_container_classes().items():
        if ctypes.lower() == db_container.type.lower():
            try:
                container = cls(db_container)
            except Exception as ex:  # pragma: no cover
                log.warning(f"Error creating container from db object: {ex}")
                return None
            return container
    return None


@log_with(log)
def find_container_by_id(container_id: int) -> TokenContainerClass:
    """
    Returns the TokenContainerClass object for the given container id or raises a ResourceNotFoundError.

    :param container_id: ID of the container
    :return: container object
    """
    db_container = TokenContainer.query.filter(TokenContainer.id == container_id).first()
    if not db_container:
        raise ResourceNotFoundError(f"Unable to find container with id {container_id}.")

    return create_container_from_db_object(db_container)


def find_container_by_serial(serial: str) -> TokenContainerClass:
    """
    Returns the TokenContainerClass object for the given container serial or raises a ResourceNotFoundError.

    :param serial: Serial of the container
    :return: container object
    :rtype: privacyidea.lib.containerclass.TokenContainerClass
    """
    if serial:
        db_container = TokenContainer.query.filter(func.upper(TokenContainer.serial) == serial.upper()).first()
    else:
        db_container = None
    if not db_container:
        raise ResourceNotFoundError(f"Unable to find container with serial {serial}.")

    return create_container_from_db_object(db_container)


def _create_container_query(user: User = None, serial: str = None, ctype: str = None, token_serial: str = None,
                            realms: list = None, template: str = None, sortby: str = 'serial',
                            sortdir: str = 'asc') -> Query:
    """
    Generates a sql query to filter containers by the given parameters.

    :param user: container owner, optional
    :param serial: container serial, optional
    :param ctype: container type, optional
    :param token_serial: serial of a token which is assigned to the container, optional
    :param realms: list of realms to filter by, optional
    :param template: The name of the template the container was created with, optional
    :param sortby: column to sort by, default is the container serial
    :param sortdir: sort direction, default is ascending
    :return: sql query
    """
    sql_query = TokenContainer.query
    if user:
        sql_query = sql_query.join(TokenContainer.owners).filter(TokenContainerOwner.user_id == user.uid)

    if serial:
        sql_query = sql_query.filter(func.upper(TokenContainer.serial) == serial.upper())

    if ctype:
        sql_query = sql_query.filter(TokenContainer.type == ctype)

    if token_serial:
        sql_query = sql_query.join(TokenContainer.tokens).filter(Token.serial == token_serial)

    if realms:
        sql_query = sql_query.join(TokenContainer.realms).filter(Realm.name.in_(realms))

    if template:
        sql_query = sql_query.join(TokenContainer.template).filter(TokenContainerTemplate.name == template)

    if isinstance(sortby, str):
        # Check that the sort column exists and convert it to a container column
        cols = TokenContainer.__table__.columns
        if sortby in cols:
            sortby = cols.get(sortby)
        else:
            log.info(f'Unknown sort column "{sortby}". Using "serial" instead.')
            sortby = TokenContainer.serial

    if sortdir == "desc":
        sql_query = sql_query.order_by(sortby.desc())
    else:
        sql_query = sql_query.order_by(sortby.asc())

    return sql_query


def get_all_containers(user: User = None, serial: str = None, ctype: str = None, token_serial: str = None,
                       realms: list = None, sortby: str = 'serial', sortdir: str = 'asc', template: str = None,
                       page: int = 0, pagesize: int = 0) -> dict:
    """
    This function is used to retrieve a container list, that can be displayed in
    the Web UI. It supports pagination if either page or pagesize is given (e.g. >0).
    Each retrieved page will also contain a "next" and a "prev", indicating
    the next or previous page. If page and pagesize are both smaller than 0, no pagination is used.
    The containers are filtered by the given parameters.

    :param user: container owner, optional
    :param serial: container serial, optional
    :param ctype: container type, optional
    :param token_serial: serial of a token which is assigned to the container, optional
    :param realms: list of realms the container is assigned to, optional
    :param sortby: column to sort by, default is the container serial
    :param sortdir: sort direction, default is ascending
    :param template: The name of the template the container was created with, optional
    :param page: The number of the page to view. Starts with 1 ;-)
    :param pagesize: The size of the page
    :returns: A dictionary with a list of containers at the key 'containers' and optionally pagination entries ('prev',
              'next', 'current', 'count')
    """
    sql_query = _create_container_query(user=user, serial=serial, ctype=ctype, token_serial=token_serial, realms=realms,
                                        template=template, sortby=sortby, sortdir=sortdir)
    ret = {}
    # Paginate if requested
    if page > 0 or pagesize > 0:
        ret = create_pagination(page, pagesize, sql_query, "containers")
    else:  # No pagination
        ret["containers"] = sql_query.all()

    container_list = [create_container_from_db_object(db_container) for db_container in ret["containers"]]
    ret["containers"] = container_list

    return ret


def create_pagination(page: int, pagesize: int, sql_query: Query,
                      object_list_key: str) -> dict:
    """
        Creates the pagination of a sql query.

        :param page: The number of the page to view. Starts with 1
        :param pagesize: The number of objects that shall be shown on one page
        :param sql_query: The sql query to paginate
        :param object_list_key: The key used in the return dictionary for the list of objects
        :return: A dictionary with pagination information and a list of database objects
    """
    ret = {}
    if page < 1:
        page = 1
    if pagesize < 1:
        pagesize = 10

    pagination = sql_query.paginate(page=page, per_page=pagesize, error_out=False)
    db_objects = pagination.items

    prev = None
    if pagination.has_prev:
        prev = page - 1
    nxt = None
    if pagination.has_next:
        nxt = page + 1

    ret["prev"] = prev
    ret["next"] = nxt
    ret["current"] = page
    ret["count"] = pagination.total
    ret[object_list_key] = db_objects
    return ret


def find_container_for_token(serial: str) -> TokenContainerClass:
    """
    Returns a TokenContainerClass object for the given token or raises a ResourceNotFoundError
    if the token does not exist.

    :param serial: Serial of the token
    :return: container object or None if the token is not in a container
    """
    container = None
    db_token = Token.query.filter(Token.serial == serial).first()
    if not db_token:
        raise ResourceNotFoundError(f"Unable to find token with serial {serial}.")
    token_id = db_token.id
    row = TokenContainerToken.query.filter(TokenContainerToken.token_id == token_id).first()
    if row:
        container_id = row.container_id
        container = find_container_by_id(container_id)
    return container


def get_container_classes() -> dict:
    """
    Returns a dictionary of all available container classes in the format: { type: class }.
    New container types have to be added here.
    """
    # className: module
    classes = {
        "TokenContainerClass": "privacyidea.lib.containerclass",
        "SmartphoneContainer": "privacyidea.lib.containers.smartphone",
        "YubikeyContainer": "privacyidea.lib.containers.yubikey"
    }

    ret = {}
    for cls, mod in classes.items():
        try:
            m = importlib.import_module(mod)
            c = getattr(m, cls)
            ret[c.get_class_type().lower()] = c
        except Exception as ex:  # pragma: no cover
            log.warning(f"Error importing module {cls}: {ex}")

    return ret


def get_container_policy_info(container_type: Union[str, None] = None):
    """
    Returns the policy info for the given container type or for all container types if no type is defined.

    :param container_type: The type of the container, optional
    :return: The policy info for the given container type or for all container types
    """
    classes = get_container_classes()
    if container_type:
        if container_type in classes.keys():
            return classes[container_type].get_container_policy_info()
        else:
            raise ResourceNotFoundError(f"Unable to find container type {container_type}.")
    else:
        ret = {}
        for container_type, container_class in classes.items():
            ret[container_type] = container_class.get_container_policy_info()
        return ret


def init_container(params: dict) -> dict:
    """
    Create a new container with the given parameters. Requires at least the type.

    :param params: The parameters for the new container as dictionary like

        ::

            {
                "type": ...,
                "description": ..., (optional)
                "container_serial": ..., (optional)
                "user": ..., Name of the user (optional)
                "realm": ..., Name of the realm (optional)
                "template": {...}, Template as dictionary (optional)
                "template_name": ..., Name of the template (optional)
            }

        To assign a user to the container, the user and realm are required.

    :return: Dictionary containing the serial of the created container and a list of init details for tokens if the
        container is created from a template

        ::

            {
                "container_serial": "CONT0001",
                "template_tokens": [{"type": "hotp", ...}, ...]
            }
    """
    ctype = params.get("type")
    if not ctype:
        raise EnrollmentError("Type parameter is required!")
    if ctype.lower() not in get_container_classes().keys():
        raise EnrollmentError(f"Type '{ctype}' is not a valid type!")

    template_dict = params.get("template")
    template_name = params.get("template_name")
    if template_dict and template_name:
        raise ParameterError("Both template and template_name are given. Choose only one!")

    desc = params.get("description") or ""
    serial = params.get("container_serial")
    if serial:
        # Check if a container with this serial already exists
        containers = get_all_containers(serial=serial)["containers"]
        if len(containers) > 0:
            raise EnrollmentError(f"Container with serial {serial} already exists!")
    else:
        serial = _gen_serial(ctype)
    db_container = TokenContainer(serial=serial, container_type=ctype.lower(), description=desc)
    db_container.save()

    container = create_container_from_db_object(db_container)

    # Creation Date
    creation_date = datetime.datetime.now(datetime.timezone.utc).isoformat(timespec="seconds")
    container.update_container_info(
        [TokenContainerInfoData(key="creation_date", value=creation_date, info_type=PI_INTERNAL)])

    # Template handling
    template_tokens = []
    if template_name:
        # Use template from db
        try:
            template = get_template_obj(template_name)
        except ResourceNotFoundError as ex:
            template = None
            log.warning(f"Template {template_name} does not exists, create container without template: {ex}")

        if template:
            if template.container_type == ctype:
                template_options = template.get_template_options_as_dict()
                template_tokens = template_options.get("tokens", [])
                container.template = template_name
            else:
                log.warning(f"Template {template_name} is not of type {ctype}, create container without template.")
    elif template_dict:
        # Use template dictionary
        if template_dict.get("container_type") == ctype:
            # check if the template was modified, otherwise save the template name
            stored_templates = get_templates_by_query(name=template_dict["name"])["templates"]
            if len(stored_templates) > 0:
                original_template = stored_templates[0]
                original_template_used = compare_template_dicts(template_dict, original_template)
                if original_template_used:
                    container.template = original_template["name"]
            template_options = template_dict.get("template_options", {})
            # tokens from template
            template_tokens = template_options.get("tokens", [])
        else:
            log.warning(f"Template {template_name} is not of type {ctype}, create container without template.")

    user = params.get("user")
    realm = params.get("realm")
    realms = []
    if user and not realm:
        log.info(f"Assigning container {container.serial} to user {user} on "
                 f"creation requires both user and realm parameters!")
    elif realm and not user:
        realms.append(realm)
        container.set_realms(realms, add=True)
    elif user and realm:
        try:
            container.add_user(User(login=user, realm=realm))
        except UserError as ex:
            log.warning(f"Error setting user for container {serial}: {ex}")

    container.set_states(['active'])

    res = {"container_serial": serial, "template_tokens": template_tokens}
    return res


def create_container_tokens_from_template(container_serial: str, template_tokens: list, request,
                                          user_role: str) -> dict:
    """
    Create tokens for the container from the given template. The token policies are checked and the enroll information
    is read from the policies for each token. The tokens owner and the enroll information are added to the request
    object to check the corresponding policies. All errors are caught and logged to be able to create the remaining
    tokens.

    :param container_serial: The serial of the container
    :param template_tokens: The template to create the tokens from as list of dictionaries where each dictionary
        contains the details for a token to be enrolled
    :param request: The request object
    :param user_role: The role of the user ('admin' or 'user')
    :return: A dictionary containing the enroll details for each created token in the format:

    ::

        {
            <token_serial>: {"serial": <token_serial>,
                             "type": <token_type>,
                             "init_params": <params used for the enrollment>, ...},
        }
    """
    container = find_container_by_serial(container_serial)

    users = container.get_users()
    if len(users) > 0:
        container_owner = users[0]
    else:
        container_owner = User()
    realms = get_container_realms(container_serial)

    init_result = {}

    # Get policies for the token
    from privacyidea.api.lib.prepolicy import (check_max_token_realm, sms_identifiers,
                                               indexedsecret_force_attribute, pushtoken_add_config,
                                               tantoken_count, papertoken_count, init_token_length_contents,
                                               init_token_defaults, check_external, check_otp_pin, encrypt_pin,
                                               init_random_pin, twostep_enrollment_parameters,
                                               twostep_enrollment_activation, enroll_pin,
                                               init_tokenlabel, check_token_init, check_max_token_user,
                                               require_description)
    from privacyidea.api.lib.postpolicy import check_verify_enrollment, save_pin_change

    # Create each token defined in the template. The template contains the enroll information for each token.
    for token_info in template_tokens:
        token = None
        user = User()
        # If the user flag is set, the token is assigned to the container owner: set the full user information in the
        # enroll information
        if token_info.get("user"):
            if container_owner:
                token_info["user"] = container_owner.login
                token_info["realm"] = container_owner.realm
                token_info["resolver"] = container_owner.resolver
            elif realms:
                token_info["realm"] = realms[0]
                del token_info["user"]
            else:
                del token_info["user"]
            user = container_owner
        elif user_role == "user" and request.User:
            # Users are always assigned to the tokens, only admins can create tokens without a user
            user = request.User
            token_info["user"] = user.login
            token_info["realm"] = user.realm
            token_info["resolver"] = user.resolver
        elif token_info.get("user") is not None:
            del token_info["user"]

        # The pre-policy decorator functions require a request object containing the enroll information.
        # Hence, we need to clear the data in the request object from the previous token and set the new enroll
        # information for the current token.
        request.all_data = {}
        request.all_data.update(token_info)

        # Pre-policy checks
        # TODO: Refactor including original uses of these functions (decorators on token init endpoint)
        try:
            check_max_token_realm(request, None)
            require_description(request, None)
            check_max_token_user(request, None)
            check_token_init(request, None)
            init_tokenlabel(request, None)
            enroll_pin(request, None)
            twostep_enrollment_activation(request, None)
            twostep_enrollment_parameters(request, None)
            init_random_pin(request, None)
            encrypt_pin(request, None)
            check_otp_pin(request, None)
            check_external(request, None)
            init_token_defaults(request, None)
            init_token_length_contents(request, None)
            papertoken_count(request, None)
            sms_identifiers(request, None)
            tantoken_count(request, None)
            pushtoken_add_config(request, None)
            indexedsecret_force_attribute(request, None)
        except Exception as ex:
            log.warning(f"Error checking pre-policies for token {token_info} created from template: {ex}")
            continue

        init_params = request.all_data
        try:
            token = init_token(init_params, user)
            init_result[token.get_serial()] = {"type": token.get_type()}
            init_result[token.get_serial()].update(token.get_init_detail(init_params, user))
            container.add_token(token)
        except Exception as ex:
            log.warning(f"Error creating token {token_info} from template: {ex}")
            if token:
                if init_result.get(token.get_serial()):
                    del init_result[token.get_serial()]
                token.delete_token()
            continue

        # Post-policy checks
        try:
            # Post-policy decorators require a response object containing the result of the token creation.
            response = send_result(True, details=init_result[token.get_serial()])
            check_verify_enrollment(request, response)
            save_pin_change(request, response)
        except Exception as ex:
            log.warning(f"Error checking post-policy for token {token_info} created from template: {ex}")
            continue

        init_result[token.get_serial()].update(response.json["detail"])
        init_result[token.get_serial()]["init_params"] = init_params

    return init_result


def add_token_to_container(container_serial: str, token_serial: str) -> bool:
    """
    Add a single token to a container. If a token is already in a container it is removed from the old container.
    Raises a ResourceNotFoundError if either the container or token does not exist.

    :param container_serial: The serial of the container
    :param token_serial: The serial of the token
    :return: True on success
    """
    container = find_container_by_serial(container_serial)

    # Get the token object
    token = get_tokens_from_serial_or_user(token_serial, None)[0]

    # Check if the token is in a container
    old_container = find_container_for_token(token_serial)

    if old_container and old_container.serial != container.serial:
        # Remove token from old container
        remove_token_from_container(old_container.serial, token_serial)
        log.info(f"Adding token {token.get_serial()} to container {container_serial}: "
                 f"Token removed from previous container {old_container.serial}.")

    res = container.add_token(token)

    return res


def add_multiple_tokens_to_container(container_serial: str, token_serials: list) -> dict:
    """
    Add the given tokens to the container with the given serial. Raises a ResourceNotFoundError if the container does
    not exist. If a token is already in a container it is removed from the old container.

    :param container_serial: The serial of the container
    :param token_serials: A list of token serials to add
    :return: A dictionary in the format {<token_serial>: <success>}
    """
    # Raises ResourceNotFound if container does not exist
    find_container_by_serial(container_serial)

    ret = {}
    for token_serial in token_serials:
        try:
            res = add_token_to_container(container_serial, token_serial)
        except Exception as ex:
            # We are catching the exception here to be able to add the remaining tokens
            log.warning(f"Error adding token {token_serial} to container {container_serial}: {ex}")
            res = False
        ret[token_serial] = res

    return ret


def add_not_authorized_tokens_result(result: dict, not_authorized_serials: list) -> dict:
    """
    Add the result False for all tokens the user is not authorized to manage.

    :param result: The result dictionary in the format {token_serial: success}
    :param not_authorized_serials: A list of token serials the user is not authorized to manage
    :return: The result dictionary with the not authorized tokens added like {<token_serial>: False}
    """
    if not_authorized_serials:
        for serial in not_authorized_serials:
            result[serial] = False
    return result


def get_container_classes_descriptions() -> dict:
    """
    Returns a dictionary of {"type": "Type: description"} entries for all container types.
    Used to list the container types.
    """
    ret = {}
    classes = get_container_classes()
    for container_type, container_class in classes.items():
        ret[container_type] = f"{container_type.capitalize()}: {container_class.get_class_description()}"
    return ret


def get_container_token_types() -> dict:
    """
    Returns a dictionary of {"type": ["tokentype0", "tokentype1", ...]} entries for all container types.
    Used to list the supported token types for each container type.
    """
    ret = {}
    classes = get_container_classes()
    for container_type, container_class in classes.items():
        ret[container_type] = container_class.get_supported_token_types()
    return ret


def remove_token_from_container(container_serial: str, token_serial: str) -> bool:
    """
    Remove the given token from the container with the given serial.
    Raises a ResourceNotFoundError if the container or token does not exist.

    :param container_serial: The serial of the container
    :param token_serial: the serial of the token to remove
    :return: True on success
    """
    container = find_container_by_serial(container_serial)
    res = container.remove_token(token_serial)

    return res


def remove_multiple_tokens_from_container(container_serial: str, token_serials: str) -> dict:
    """
    Remove the given tokens from the container with the given serial.
    Raises a ResourceNotFoundError if no container for the given serial exist.
    Errors of removing tokens are caught and only logged, in order to be able to remove the remaining
    tokens in the list.

    :param container_serial: The serial of the container
    :param token_serials: A list of token serials to remove
    :return: A dictionary in the format {token_serial: success}
    """
    # Check that container exists
    find_container_by_serial(container_serial)

    ret = {}
    for token_serial in token_serials:
        try:
            res = remove_token_from_container(container_serial, token_serial)
        except Exception as ex:
            # We are catching the exception here to be able to remove the remaining tokens
            log.warning(f"Error removing token {token_serial} from container {container_serial}: {ex}")
            res = False
        ret[token_serial] = res
    return ret


def add_container_info(serial: str, ikey: str, ivalue) -> bool:
    """
    Add the given info to the container with the given serial.
    If the key already exists, the value is updated. However, if the entry is of type PI_INTERNAL, the value can not be
    modified.

    :param serial: The serial of the container
    :param ikey: The info key
    :param ivalue: The info value
    :returns: True on success
    """
    container = find_container_by_serial(serial)

    # Check if key already exists and if it is an internal key
    internal_keys = container.get_internal_info_keys()
    if ikey in internal_keys:
        raise PolicyError(f"The key {ikey} is an internal entry and can not be modified.")

    container.update_container_info([TokenContainerInfoData(key=ikey, value=ivalue)])
    return True


def set_container_info(serial, info: dict) -> dict:
    """
    Set the given info to the container with the given serial.
    Keys of type PI_INTERNAL can not be modified and will be ignored.

    :param serial: The serial of the container
    :param info: The info dictionary in the format {key: value}
    :returns: Dictionary with the success state for each info key
    """
    container = find_container_by_serial(serial)
    result = {}

    # Remove internal keys from the info dictionary, they can not be modified by the user
    internal_keys = container.get_internal_info_keys()
    not_internal_info = {}
    for key, value in info.items():
        if key not in internal_keys:
            not_internal_info[key] = value
            result[key] = True
        else:
            result[key] = False
            log.warning(f"The key {key} is an internal entry and can not be modified.")

    container.set_container_info(not_internal_info)
    return result


def get_container_info_dict(serial: str, ikey: str = None) -> dict:
    """
    Returns the info of the given key or all infos if no key is given for the container with the given serial.

    :param serial: The serial of the container
    :param ikey: The info key or None to get all info keys
    :return: The info dict
    """
    container = find_container_by_serial(serial)

    container_info = {container_info.key: container_info.value for container_info in container.get_container_info()}
    if ikey:
        if ikey in container_info.keys():
            container_info = {ikey: container_info[ikey]}
        else:
            container_info = {ikey: None}
            log.warning(f"Info key {ikey} not found in container {serial}.")
    return container_info


def delete_container_info(serial: str, ikey: str = None) -> dict:
    """
    Delete the info of the given key or all infos if no key is given.
    Internal infos are not deleted

    :param serial: The serial of the container
    :param ikey: The info key or None to delete all info keys
    :return: Dictionary with all info keys or only ikey if given and the value True on success, False otherwise
    """
    container = find_container_by_serial(serial)

    res = container.delete_container_info(ikey, keep_internal=True)
    return res


def assign_user(serial: str, user: User) -> bool:
    """
    Assign a user to a container.

    :param serial: container serial
    :param user: user to assign to the container
    :return: True on success, False otherwise
    """
    container = find_container_by_serial(serial)

    res = container.add_user(user)
    return res


def unassign_user(serial: str, user: User) -> bool:
    """
    Unassign a user from a container.

    :param serial: container serial
    :param user: user to unassign from the container
    :return: True on success, False otherwise
    """
    container = find_container_by_serial(serial)

    res = container.remove_user(user)
    return res


def set_container_description(serial: str, description: str):
    """
    Set the description of a container.

    :param serial: serial of the container
    :param description: new description
    """
    container = find_container_by_serial(serial)

    container.description = description


def set_container_states(serial: str, states: list) -> dict:
    """
    Set the states of a container.

    :param serial: serial of the container
    :param states: new states as list of str
    :returns: Dictionary in the format {state: success}
    """
    container = find_container_by_serial(serial)

    res = container.set_states(states)
    return res


def add_container_states(serial: str, states: list) -> dict:
    """
    Add the states to a container.

    :param serial: serial of the container
    :param states: additional states as list of str
    :returns: Dictionary in the format {state: success}
    """
    container = find_container_by_serial(serial)

    res = container.add_states(states)
    return res


def set_container_realms(serial: str, realms: list,
                         allowed_realms: Union[list, None] = []) -> dict:
    """
    Set the realms of a container.

    :param serial: serial of the container
    :param realms: new realms as list of str
    :param allowed_realms: A list of realms the admin is allowed to set (None if all realms are allowed), optional
    :returns: Dictionary in the format {realm: success}, the entry 'deleted' indicates whether existing realms were
              deleted.
    """
    container = find_container_by_serial(serial)
    old_realms = [realm.name for realm in container.realms]

    # Check if admin is allowed to set the realms
    matching_realms = realms
    res_failed = {}
    if allowed_realms:
        matching_realms = list(set(realms).intersection(allowed_realms))
        excluded_realms = list(set(realms) - set(matching_realms))
        if len(excluded_realms) > 0:
            log.info(f"User is not allowed to set realms {excluded_realms} for container {serial}.")
            res_failed = {realm: False for realm in excluded_realms}

        # Check if admin is allowed to remove the old realms
        not_allowed_realms = set(old_realms) - set(allowed_realms)
        # Add realms that are not allowed to be removed to the set list
        matching_realms = list(set(matching_realms).union(not_allowed_realms))

    # Set realms
    res = container.set_realms(matching_realms, add=False)
    res.update(res_failed)
    return res


def add_container_realms(serial: str, realms: list, allowed_realms: Union[list, None]) -> dict:
    """
    Add the realms to the container realms.

    :param serial: serial of the container
    :param realms: new realms as list of str
    :param allowed_realms: A list of realms the admin is allowed to set, optional
    :returns: Dictionary in the format {realm: success}, the entry 'deleted' indicates whether existing realms were
              deleted.
    """
    container = find_container_by_serial(serial)

    # Check if admin is allowed to set the realms
    matching_realms = realms
    res_failed = {}
    if allowed_realms:
        matching_realms = list(set(realms).intersection(allowed_realms))
        excluded_realms = list(set(realms) - set(matching_realms))
        if len(excluded_realms) > 0:
            log.info(f"User is not allowed to set realms {excluded_realms} for container {serial}.")
            res_failed = {realm: False for realm in excluded_realms}

    # Add realms
    res = container.set_realms(matching_realms, add=True)
    res.update(res_failed)
    return res


def get_container_realms(serial: str) -> list:
    """
    Get the realms of the container.

    :param serial: serial of the container
    :returns: List of realm names
    """
    container = find_container_by_serial(serial)
    return [realm.name for realm in container.realms]


def create_container_dict(container_list: list, no_token: bool = False, user: User = None,
                          logged_in_user_role: str = 'user', allowed_token_realms: Union[list, None] = [],
                          hide_token_info: list = None) -> list:
    """
    Create a dictionary for each container in the list.
    It contains the container properties, owners, realms, tokens and info.
    The information is only provided if the user is allowed to see it.

    :param container_list: List of container objects
    :param no_token: If True, the token information is not included
    :param user: The user object requesting the containers
    :param logged_in_user_role: The role of the logged-in user ('admin' or 'user')
    :param allowed_token_realms: A list of realms the admin is allowed to see tokens from
    :param hide_token_info: List of token info keys to hide in the response, optional
    :return: List of container dictionaries

    Example of a returned list:
        ::

            [
                {
                    "type": "generic",
                    "serial": "CONT0001",
                    "description": "Container description",
                    "last_authentication": "2021-06-01T12:00:00+00:00",
                    "last_synchronization": "2021-06-01T12:00:00+00:00",
                    "states": ["active"],
                    "users": [
                        {
                            "user_name": "user1",
                            "user_realm": "realm1",
                            "user_resolver": "resolver1",
                            "user_id": 1
                        }
                        ],
                    "tokens": [
                        {
                            "serial": "TOTP0001",
                            "type": "totp",
                            "active": true,
                            ...
                        }],
                    "info": {"hash_algorithm": "SHA256", ...},
                    "internal_info_keys": ["hash_algorithm"],
                    "realms": ["realm1", "realm2"],
                    "template": "template1"
                }, ...
            ]
    """
    res: list = []
    for container in container_list:
        container_dict = container.get_as_dict(include_tokens=not no_token, public_info=True)
        if not no_token:
            token_serials = ",".join(container_dict["tokens"])
            tokens_dict_list = []
            if len(token_serials) > 0:
                tokens = get_tokens(serial=token_serials)
                tokens_dict_list = convert_token_objects_to_dicts(tokens, user=user, user_role=logged_in_user_role,
                                                                  allowed_realms=allowed_token_realms,
                                                                  hidden_token_info=hide_token_info)
            container_dict["tokens"] = tokens_dict_list

        res.append(container_dict)

    return res


def create_endpoint_url(base_url: str, endpoint: str) -> str:
    """
    Creates the url for an endpoint. It concat the base_url and the endpoint if the endpoint is not already in the
    base_url. base_url and endpoint are separated by a slash.

    :param base_url: The base url of the host
    :param endpoint: The endpoint
    :return: The url for the endpoint
    :rtype: str
    """
    if endpoint not in base_url:
        if base_url[-1] != "/":
            base_url += "/"
        endpoint_url = base_url + endpoint
    else:
        endpoint_url = base_url
    return endpoint_url


def finalize_registration(container_serial: str, params: dict) -> dict:
    """
    Finalize the registration of a container if the challenge response is valid.
    If the container is in the registration_state `rollover`, it finalizes the container rollover.

    :param container_serial: The serial of the container
    :param params: The parameters for the registration as dictionary
    :return: dictionary with container specific information
    """
    # Get container
    container = find_container_by_serial(container_serial)
    container_info = container.get_container_info_dict()
    registration_state = container_info.get("registration_state")

    # Update params with registration url
    if registration_state == "rollover":
        server_url = container_info.get("rollover_server_url")
    else:
        server_url = container_info.get("server_url")
    if server_url is None:
        log.debug("Server url is not set in the container info. Ensure that registration/init is called first.")
        server_url = " "
    scope = create_endpoint_url(server_url, "container/register/finalize")
    params.update({'scope': scope})

    res = container.finalize_registration(params)

    if registration_state == "rollover":
        # container registration rolled over: set rollover info as correct info
        for key, value in container_info.items():
            if key.find("rollover_") == 0:
                original_key = key.replace("rollover_", "")
                container.update_container_info(
                    [TokenContainerInfoData(key=original_key, value=value, info_type=PI_INTERNAL)])
                container.delete_container_info(key, keep_internal=False)

        finalize_container_rollover(container)
        container.update_container_info(
            [TokenContainerInfoData(key="registration_state", value="rollover_completed", info_type=PI_INTERNAL)])

    return res


def finalize_container_rollover(container: TokenContainerClass):
    """
    Finalize the rollover of a container. For each token in the container a rollover is performed.
    All previous challenges are deleted.

    :param container: The container object
    """

    tokens = container.get_tokens()

    # Offline tokens can not be rolled over, that would invalidate the offline otp values
    offline_serials = [token.get_serial() for token in tokens if is_offline_token(token.get_serial())]
    online_tokens = [token for token in tokens if token.get_serial() not in offline_serials]
    if len(offline_serials) > 0:
        log.info(f"The following offline tokens are in the container: {offline_serials}. "
                 "They can not be rolled over.")

    for token in online_tokens:
        params = {"serial": token.get_serial(),
                  "type": token.get_type(),
                  "genkey": True,
                  "rollover": True}
        token_info = token.get_tokeninfo()
        params.update(token_info)
        try:
            token = init_token(params)
        except Exception as ex:
            # Do not block the rollover process
            log.debug(f"Error during rollover of token {token.get_serial()} in container rollover: {ex}")

    # Delete previous challenges of the container
    delete_challenges(container.serial)


def init_container_rollover(container: TokenContainerClass, server_url: str, challenge_ttl: int, registration_ttl: int,
                            ssl_verify: bool, params: dict) -> dict:
    """
    Initializes the rollover of a container.
    First the response to the challenge is validated. If it is valid, the registration is initialized.
    The new registration info is not finally set until the new container successfully finalized the registration.

    :param container: The container object
    :param server_url: The server url of the privacyIDEA server the client can contact
    :param challenge_ttl: The time to live of the challenge in minutes
    :param registration_ttl: The time to live of the challenge for the registration in minutes
    :param ssl_verify: If the client has to verify the ssl certificate of the server
    :param params: Container type specific parameters for the registration as dictionary
    :return: dictionary with container specific information for the client
    """
    # Check challenge if rollover is allowed
    rollover_scope = create_endpoint_url(server_url, "container/rollover")
    params.update({"scope": rollover_scope})
    container.check_challenge_response(params)

    registration_scope = create_endpoint_url(server_url, "container/register/finalize")
    params.update({"scope": registration_scope})

    # Get registration data
    res = container.init_registration(server_url, registration_scope, registration_ttl, ssl_verify, params)

    # Set registration state
    info = [TokenContainerInfoData(key="registration_state", value="rollover", info_type=PI_INTERNAL),
            TokenContainerInfoData(key="rollover_server_url", value=server_url, info_type=PI_INTERNAL),
            TokenContainerInfoData(key="rollover_challenge_ttl", value=str(challenge_ttl), info_type=PI_INTERNAL)]
    container.update_container_info(info)

    return res


def unregister(container: TokenContainerClass) -> bool:
    """
    Unregister a container from the synchronization and deletes all challenges for the container.

    :param container: The container object
    :return: True on success
    """
    # terminate registration
    container.terminate_registration()

    # Delete all challenges of the container
    delete_challenges(serial=container.serial)

    return True


def set_options(serial: str, options: dict):
    """
    Set the options of a container. The user has to be an admin or the owner of the container.

    :param serial: The serial of the container
    :param options: The options as dictionary
    """
    container = find_container_by_serial(serial)

    container.add_options(options)


def get_container_template_classes() -> dict:
    """
    Returns a dictionary of all available container template classes in the format: { type: class }.
    New container template types have to be added here.
    """
    # className: module
    classes = {
        "ContainerTemplateBase": "privacyidea.lib.containertemplate.containertemplatebase",
        "SmartphoneContainerTemplate": "privacyidea.lib.containertemplate.smartphonetemplate",
        "YubikeyContainerTemplate": "privacyidea.lib.containertemplate.yubikeytemplate"
    }

    ret = {}
    for cls, mod in classes.items():
        try:
            m = importlib.import_module(mod)
            c = getattr(m, cls)
            ret[c.get_class_type().lower()] = c
        except Exception as ex:  # pragma: no cover
            log.warning(f"Error importing module {cls}: {ex}")

    return ret


def create_container_template(container_type: str, template_name: str, options: dict, default: bool = False) -> int:
    """
    Create a new container template.

    :param container_type: The type of the container
    :param template_name: The name of the template
    :param options: The options for the template as dictionary
    :param default: If True, the template is set as default, optional

    Example for the options dictionary:
        ::

            {
                "tokens": [{"type": "hotp", "genkey": True, "hashlib": "sha256"}, ...]
            }

    :return: ID of the created template
    """
    # Check container type
    if container_type.lower() not in get_container_classes().keys():
        raise EnrollmentError(f"Type '{container_type}' is not a valid type!")

    TokenContainerTemplate(name=template_name, container_type=container_type).save()
    template = get_template_obj(template_name)
    try:
        if options:
            template.template_options = options
        if default:
            template.default = default
    except Exception as ex:
        # We need to delete the template on error, but still want to raise the original exception
        template.delete()
        raise ex

    return template.id


def create_container_template_from_db_object(db_template: TokenContainerTemplate) -> Union[ContainerTemplateBase, None]:
    """
    Create a TokenContainerTemplate object from the given db object.

    :param db_template: The DB object to create the container template from
    :return: The created container template object or None if the container template type is not supported
    """

    for ctypes, cls in get_container_template_classes().items():
        if ctypes.lower() == db_template.container_type.lower():
            try:
                template = cls(db_template)
            except Exception as ex:  # pragma: no cover
                log.warning(f"Error creating container template from db object: {ex}")
                return None
            return template
    return None


def get_all_templates_with_type():
    """
    Returns a list of display strings containing the name and type of all templates.
    """
    templates = TokenContainerTemplate.query.all()
    template_list = []
    for template in templates:
        template_list.append(f"{template.name}({template.container_type})")
    return template_list


def get_templates_by_query(name: str = None, container_type: str = None, default: bool = None, page: int = 0,
                           pagesize: int = 0, sortdir: str = "asc",
                           sortby: str = "name") -> dict:
    """
    Returns a list of all templates or a list filtered by the given parameters.

    :param name: The name of the template, optional
    :param container_type: The type of the container, optional
    :param default: Filters for default templates if True or non-default if False, optional
    :param page: The number of the page to view. 0 if no pagination shall be used
    :param pagesize: The size of the page. 0 if no pagination shall be used
    :param sortdir: The sort direction, either 'asc' or 'desc'
    :param sortby: The attribute to sort by
    :return: a dictionary with a list of templates at the key 'templates' and optionally pagination entries ('prev',
             'next', 'current', 'count')
    """
    sql_query = TokenContainerTemplate.query
    if name:
        sql_query = sql_query.filter(TokenContainerTemplate.name == name)
    if container_type:
        sql_query = sql_query.filter(TokenContainerTemplate.container_type == container_type)
    if default is not None:
        sql_query = sql_query.filter(TokenContainerTemplate.default == default)

    if isinstance(sortby, str):
        # Check that the sort column exists and convert it to a template column
        cols = TokenContainerTemplate.__table__.columns
        if sortby in cols:
            sortby = cols.get(sortby)
        else:
            log.info(f'Unknown sort column "{sortby}". Using "name" instead.')
            sortby = TokenContainerTemplate.name

    if sortdir == "desc":
        sql_query = sql_query.order_by(sortby.desc())
    else:
        sql_query = sql_query.order_by(sortby.asc())

    # paginate if requested
    if page > 0 or pagesize > 0:
        ret = create_pagination(page, pagesize, sql_query, "templates")
    else:
        ret = {"templates": sql_query.all()}

    # create class objects from db objects
    template_obj_list = [create_container_template_from_db_object(template) for template in ret["templates"]]

    # convert to dict
    template_list = []
    for template in template_obj_list:
        template_options = {}
        if template.template_options != "":
            template_options = json.loads(template.template_options)
        template_dict = {"name": template.name,
                         "container_type": template.container_type,
                         "template_options": template_options,
                         "default": template.default}
        template_list.append(template_dict)

    ret["templates"] = template_list

    return ret


def get_template_obj(template_name: str) -> ContainerTemplateBase:
    """
    Returns the template class object for the given template name.
    Raises a ResourceNotFoundError if no template with this name exists.

    :param template_name: The name of the template
    :return: The template class object
    """
    db_template = TokenContainerTemplate.query.filter(TokenContainerTemplate.name == template_name).first()
    if not db_template:
        raise ResourceNotFoundError(f"Template {template_name} does not exist.")
    template = create_container_template_from_db_object(db_template)
    return template


def set_default_template(name: str):
    """
    Sets the template of the given name as default and all other templates for the container type as non-default.

    :param name: The name of the template to be the new default template
    """
    default_template = get_template_obj(name)

    # Get all default templates for the container type and reset them to non-default
    old_default_templates = get_templates_by_query(container_type=default_template.container_type, default=True)
    for template in old_default_templates["templates"]:
        template_obj = get_template_obj(template["name"])
        template_obj.default = False

    default_template.default = True


def compare_template_dicts(template_a: dict, template_b: dict) -> bool:
    """
    Compares two template dictionaries for equal tokens.

    :param template_a: The first template dictionary
    :param template_b: The second template dictionary
    :return: True if the templates contain the same tokens, False otherwise.
    """
    if template_a is None or template_b is None:
        return False

    # get template options
    template_options_a = template_a.get("template_options", {})
    template_options_b = template_b.get("template_options", {})

    # compare tokens
    tokens_a = template_options_a.get("tokens", [])
    tokens_b = template_options_b.get("tokens", [])
    if len(tokens_a) != len(tokens_b):
        # different number of tokens, templates can not be equal
        return False

    unique_tokens_a = [token for token in tokens_a if token not in tokens_b]
    unique_tokens_b = [token for token in tokens_b if token not in tokens_a]
    if len(unique_tokens_a) > 0 or len(unique_tokens_b) > 0:
        return False

    return True


def compare_template_with_container(template: ContainerTemplateBase, container: TokenContainerClass) -> dict:
    """
    Compares the template with the container. It is only evaluated if the token types are equal.

    :param template: The template object
    :param container: The container object
    :return: A dictionary with the differences between the template and the container

    Example of a returned dictionary:
        ::

            {
                "tokens": {
                            "missing": ["hotp"],
                            "additional": ["totp"]
                            }
            }
    """
    result = {"tokens": {"missing": [], "additional": []}}
    template_options = json.loads(template.template_options)

    # compare tokens
    template_tokens = template_options.get("tokens", [])
    template_token_types = [token["type"] for token in template_tokens]
    template_token_count = {ttype: template_token_types.count(ttype) for ttype in template_token_types}
    container_token_types = [token.type for token in container.get_tokens()]
    container_token_count = {ttype: container_token_types.count(ttype) for ttype in container_token_types}

    for ttype, count_template in template_token_count.items():
        count_container = container_token_count.get(ttype, 0)
        if count_template > count_container:
            result["tokens"]["missing"].extend([ttype] * (count_template - count_container))

    for ttype, count_container in container_token_count.items():
        count_template = template_token_count.get(ttype, 0)
        if count_template < count_container:
            result["tokens"]["additional"].extend([ttype] * (count_container - count_template))

    # Check if container and template are equal
    if len(result["tokens"]["missing"]) == 0 and len(result["tokens"]["additional"]) == 0:
        result["tokens"]["equal"] = True
    else:
        result["tokens"]["equal"] = False

    return result


def get_offline_token_serials(container: TokenContainerClass) -> list:
    """
    Returns a list of serials of offline tokens in the container.

    :param container: A TokenContainerClass object
    :return: List of serials of offline tokens in the container
    """
    tokens = container.get_tokens()
    offline_serials = [token.get_serial() for token in tokens if is_offline_token(token.get_serial())]
    return offline_serials



#!/usr/bin/env python
#  2020-04-28 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Read tables oc_accounts and oc_users from owncloud
#
__doc__ = """You can use this script to read the tables oc_accounts and
oc_users from owncloud and fill a local user table in privacyIDEA.

Run this script in a cron job. It will read the users from ownCloud and
* insert new users
* update changed users
* remove deleted users
"""
from sqlalchemy import create_engine
from sqlalchemy.sql import select
from sqlalchemy.schema import ForeignKey
from sqlalchemy import (Table, MetaData, Column, Integer, Unicode)
import sys
import json
import getopt


EXAMPLE_CONFIG_FILE = """{
    "SQL": {
        "OWNCLOUD_URI": "mysql+pymysql://oc:password@localhost/oc",
        "PRIVACYIDEA_URI": "mysql+pymysql://pi:password@localhost/pi?charset=utf8mb4",
        "LOCAL_TABLE": "pi_oc_users",
        "INSERT_CHUNK_SIZE": 10000
    }
}"""


class Config(object):

    def __init__(self, config_file):
        with open(config_file, "r") as f:
            contents = f.read()
        config = json.loads(contents)
        self.OWNCLOUD_URI = config.get("SQL").get("OWNCLOUD_URI")
        self.PRIVACYIDEA_URI = config.get("SQL").get("PRIVACYIDEA_URI")
        self.LOCAL_TABLE = config.get("SQL").get("LOCAL_TABLE")
        self.INSERT_CHUNK_SIZE = config.get("SQL").get("INSERT_CHUNK_SIZE")


def sync_owncloud(config_obj):

    metadata = MetaData()

    user_table = Table(config_obj.LOCAL_TABLE, metadata,
                       Column("id", Integer, primary_key=True, nullable=False),
                       Column("email", Unicode(255), nullable=True),
                       Column("user_id", Unicode(255), nullable=False, unique=True),
                       Column("lower_user_id", Unicode(255), nullable=False, unique=True),
                       Column("display_name", Unicode(255)),
                       Column("backend", Unicode(64)),
                       Column("last_login", Integer, default=0),
                       Column("state", Integer, default=0),
                       Column("password", Unicode(255), nullable=False)
                       )

    oc_accounts_table = Table("oc_accounts", metadata,
                              Column("id", Integer, primary_key=True, nullable=False),
                              Column("email", Unicode(255), nullable=True),
                              Column("user_id", Unicode(255), nullable=False, unique=True),
                              Column("lower_user_id", Unicode(255), nullable=False, unique=True),
                              Column("display_name", Unicode(255)),
                              Column("backend", Unicode(64)),
                              Column("last_login", Integer, default=0),
                              Column("state", Integer, default=0)
                              )

    oc_users_table = Table("oc_users", metadata,
                           Column("uid", Unicode(255), ForeignKey("oc_accounts.user_id")),
                           Column("password", Unicode(255), nullable=False)
                           )

    oc_engine = create_engine(config_obj.OWNCLOUD_URI)
    privacyidea_engine = create_engine(config_obj.PRIVACYIDEA_URI)

    print("Creating table {0!s}, if it does not exist.".format(config_obj.LOCAL_TABLE))
    metadata.create_all(privacyidea_engine)

    conn_oc = oc_engine.connect()
    conn_pi = privacyidea_engine.connect()

    def insert_chunks(conn, table, values, chunk_size=100000):
        """
        Split **values** into chunks of size **chunk_size** and insert them sequentially.
        """
        values_length = len(values)
        for chunk in range(0, values_length, chunk_size):
            print('Insert records {} to {} ...'.format(chunk, min(chunk + chunk_size,
                                                                  values_length) - 1))
            try:
                conn.execute(table.insert(), values[chunk:chunk + chunk_size])
            except Exception as err:
                t = 'Failed to insert chunk: {0!s}'.format(err)
                warnings.append(t)
                print(t)

    warnings = []

    s = select([oc_accounts_table, oc_users_table.c.password]).select_from(
        oc_accounts_table.join(oc_users_table, oc_users_table.c.uid == oc_accounts_table.c.user_id))

    owncloud_source = conn_oc.execute(s)

    s = select([user_table])

    privacyidea_dest = conn_pi.execute(s)

    # Build a dict with the existing users
    pi_users = {}
    for r in privacyidea_dest:
        pi_users[r.id] = r

    pi_users_insert = []
    pi_users_update = []
    unchanged = 0
    for r in owncloud_source:
        if r.id not in pi_users.keys():
            # This is a new entry
            pi_users_insert.append(dict(id=r.id, email=r.email, user_id=r.user_id,
                                        lower_user_id=r.lower_user_id, display_name=r.display_name, password=r.password,
                                        backend=r.backend, last_login=r.last_login, state=r.state))
        else:
            # This is an existing entry
            # Check if the entry is the same
            if r == pi_users[r.id]:
                # The values are the same
                print("Entry {0!s}/{1!s} unchanged.".format(r.id, r.user_id))
                unchanged += 1
            else:
                # add to update
                pi_users_update.append(dict(id=r.id, email=r.email, user_id=r.user_id,
                                        lower_user_id=r.lower_user_id, display_name=r.display_name, password=r.password,
                                        backend=r.backend, last_login=r.last_login, state=r.state))
            # Delete entry from the privacyIDEA user list
            del(pi_users[r.id])

    pi_users_delete = pi_users

    print("Processing...")
    print("{0!s} new entries.".format(len(pi_users_insert)))
    print("{0!s} unchanged entries.".format(unchanged))
    print("{0!s} updated entries.".format(len(pi_users_update)))
    print("{0!s} removed entries.".format(len(pi_users_delete)))

    if len(pi_users_insert):
        print("Inserting new entries.")
        insert_chunks(conn_pi, user_table, pi_users_insert, config_obj.INSERT_CHUNK_SIZE)

    if len(pi_users_update):
        print("Updating entries.")
        for upd in pi_users_update:
            stmt = user_table.update().where(user_table.c.id == upd.get("id")).values(upd)
            conn_pi.execute(stmt)

    if len(pi_users_delete):
        print("Deleting removed entries.")
        for udel in pi_users_delete:
            stmt = user_table.delete().where(user_table.c.id == udel)
            conn_pi.execute(stmt)

    if warnings:
        print("We need to inform you about the following WARNINGS:")
        for warning in warnings:
            print(warning)


def usage():
    print("""
privacyidea-sync-owncloud.py --generate-example-config [--config <config file>]

    --generate-example-config, -g   Output an example config file.
                                    This is a JSON file, that needs to be passed
                                    to this command.

    --config, -c <file>             The config file, that contains the complete
                                    configuration.

{0!s}""".format(__doc__))


def main():
    try:
        opts, args = getopt.getopt(sys.argv[1:], "gc:", ["generate-example-config", "config="])
    except getopt.GetoptError as e:
        print(str(e))
        sys.exit(1)

    config_file = None
    generate_config = False

    for o, a in opts:
        if o in ("-g", "--generate-example-config"):
            generate_config = True
            print(EXAMPLE_CONFIG_FILE)
        elif o in ("-c", "--config"):
            config_file = a
        else:
            print("Unknown parameter: {0!s}".format(o))
            sys.exit(3)

    if config_file:
        config_obj = Config(config_file)
        sync_owncloud(config_obj)
        sys.exit(0)

    else:
        if not generate_config:
            usage()
            sys.exit(1)


if __name__ == '__main__':
    main()
